'''
Classes for BMI decoding using the Kalman filter. 
'''

import numpy as np
from scipy.io import loadmat

from . import bmi

class KalmanFilter(bmi.GaussianStateHMM):
    """
    Low-level KF, agnostic to application

    Model: 
       x_{t+1} = Ax_t + w_t;   w_t ~ N(0, W)
           y_t = Cx_t + q_t;   q_t ~ N(0, Q)
    """
    model_attrs = ['A', 'W', 'C', 'Q', 'C_xpose_Q_inv', 'C_xpose_Q_inv_C']
    attrs_to_pickle = ['A', 'W', 'C', 'Q', 'C_xpose_Q_inv', 'C_xpose_Q_inv_C', 'R', 'S', 'T', 'ESS']

    def __init__(self, A=None, W=None, C=None, Q=None, is_stochastic=None):
        '''
        Constructor for KalmanFilter    

        Parameters
        ----------
        A : np.mat, optional
            Model of state transition matrix
        W : np.mat, optional
            Model of process noise covariance
        C : np.mat, optional
            Model of conditional distribution between observations and hidden state
        Q : np.mat, optional
            Model of observation noise covariance
        is_stochastic : np.array, optional
            Array of booleans specifying for each state whether it is stochastic. 
            If 'None' specified, all states are assumed to be stochastic

        Returns
        -------
        KalmanFilter instance
        '''
        if A is None and W is None and C is None and Q is None:
            ## This condition should only be true in the unpickling phase
            pass
        else:
            self.A = np.mat(A)
            self.W = np.mat(W)
            self.C = np.mat(C)
            self.Q = np.mat(Q)

            if is_stochastic is None:
                n_states = self.A.shape[0]
                self.is_stochastic = np.ones(n_states, dtype=bool)
            else:
                self.is_stochastic = is_stochastic
            
            self.state_noise = bmi.GaussianState(0.0, self.W)
            self.obs_noise = bmi.GaussianState(0.0, self.Q)
            self._pickle_init()

    def _pickle_init(self):
        """Code common to unpickling and initialization
        """
        nS = self.A.shape[0]
        offset_row = np.zeros(nS)
        offset_row[-1] = 1
        self.include_offset = np.array_equal(np.array(self.A)[-1, :], offset_row)

        self.alt = nS < self.C.shape[0] # No. of states less than no. of observations
        attrs = list(self.__dict__.keys())
        if not 'C_xpose_Q_inv_C' in attrs:
            C, Q = self.C, self.Q 
            self.C_xpose_Q_inv = C.T * np.linalg.pinv(Q)
            self.C_xpose_Q_inv_C = C.T * np.linalg.pinv(Q) * C

        try:
            self.is_stochastic
        except:
            n_states = self.A.shape[0]
            self.is_stochastic = np.ones(n_states, dtype=bool)

    def _obs_prob(self, state):
        '''
        Predict the observations based on the model parameters:
            y_est = C*x_t + Q

        Parameters
        ----------
        state : bmi.GaussianState instance
            The model-predicted state

        Returns
        -------
        bmi.GaussianState instance
            the model-predicted observations
        '''
        return self.C * state + self.obs_noise

    def _forward_infer(self, st, obs_t, Bu=None, u=None, x_target=None, F=None, obs_is_control_independent=True, **kwargs):
        '''
        Estimate p(x_t | ..., y_{t-1}, y_t)

        Parameters
        ----------
        st : GaussianState
            Current estimate (mean and cov) of hidden state
        obs_t : np.mat of shape (N, 1)
             ARG_DESCR
        Bu : DATA_TYPE, optional, default=None
             ARG_DESCR
        u : DATA_TYPE, optional, default=None
             ARG_DESCR
        x_target : DATA_TYPE, optional, default=None
             ARG_DESCR
        obs_is_control_independent : bool, optional, default=True
             ARG_DESCR
        kwargs : optional kwargs
            ARG_DESCR

        Returns
        -------
        GaussianState
            New state estimate incorporating the most recent observation

        '''
        using_control_input = (Bu is not None) or (u is not None) or (x_target is not None)
        pred_state = self._ssm_pred(st, target_state=x_target, Bu=Bu, u=u, F=F)

        C, Q = self.C, self.Q
        P = pred_state.cov

        K = self._calc_kalman_gain(P)
        I = np.mat(np.eye(self.C.shape[1]))
        D = self.C_xpose_Q_inv_C
        KC = P*(I - D*P*(I + D*P).I)*D
        F = (I - KC)*self.A

        post_state = pred_state

        if obs_is_control_independent and using_control_input:
            post_state.mean += -KC*self.A*st.mean + K*obs_t
        else:
            post_state.mean += -KC*pred_state.mean + K*obs_t

        post_state.cov = (I - KC) * P 

        return post_state

    def set_state_cov(self, n_steps):
        C, Q = self.C, self.Q
        A, W = self.A, self.W
        P = self.state.cov
        for k in range(n_steps):
            
            P = A*P*A.T + W

            K = self._calc_kalman_gain(P)
            I = np.mat(np.eye(self.C.shape[1]))
            D = self.C_xpose_Q_inv_C
            KC = P*(I - D*P*(I + D*P).I)*D
            P = (I - KC) * P 

        return P

    def _calc_kalman_gain(self, P):
        '''
        Calculate Kalman gain using the 'alternate' definition

        Parameters
        ----------
        P : np.matrix
            Prediciton covariance matrix, i.e., cov(x_{t+1} | y_1, \cdots, y_t)

        Returns
        -------
        K : np.matrix
            Kalman gain matrix for the input next state prediciton covariance.        
        '''
        nX = P.shape[0]
        I = np.mat(np.eye(nX))
        D = self.C_xpose_Q_inv_C
        L = self.C_xpose_Q_inv
        K = P * (I - D*P*(I + D*P).I) * L
        return K

    def get_sskf(self, tol=1e-15, return_P=False, dtype=np.array, max_iter=4000,
        verbose=False, return_Khist=False, alt=True):
        """Calculate the steady-state KF matrices

        value of P returned is the posterior error cov, i.e. P_{t|t}

        Parameters
        ----------

        Returns
        -------        
        """ 
        A, W, C, Q = np.mat(self.A), np.mat(self.W), np.mat(self.C), np.mat(self.Q)

        nS = A.shape[0]
        P = np.mat(np.zeros([nS, nS]))
        I = np.mat(np.eye(nS))

        D = self.C_xpose_Q_inv_C 

        last_K = np.mat(np.ones(C.T.shape))*np.inf
        K = np.mat(np.ones(C.T.shape))*0

        K_hist = []

        iter_idx = 0
        last_P = None
        while np.linalg.norm(K-last_K) > tol and iter_idx < max_iter:
            P = A*P*A.T + W 
            last_K = K
            K = self._calc_kalman_gain(P)
            K_hist.append(K)
            KC = P*(I - D*P*(I + D*P).I)*D
            last_P = P
            P -= KC*P;
            iter_idx += 1
        if verbose: 
            print(("Converged in %d iterations--error: %g" % (iter_idx, np.linalg.norm(K-last_K)))) 
    
        n_state_vars, n_state_vars = A.shape
        F = (np.mat(np.eye(n_state_vars, n_state_vars)) - KC) * A
    
        if return_P and return_Khist:
            return dtype(F), dtype(K), dtype(last_P), K_hist
        elif return_P:
            return dtype(F), dtype(K), dtype(last_P)
        elif return_Khist:
            return dtype(F), dtype(K), K_hist
        else:
            return dtype(F), dtype(K)


    def get_kalman_gain_seq(self, N=1000, tol=1e-10, verbose=False):
        '''
        Calculate K_t for times {0, 1, ..., N}

        Parameters
        ----------
        N : int, optional
            Number of steps to calculate Kalman gain for, default = 1000
        tol : float, optional
            Tolerance on K matrix convergence, default = 1e-10
        verbose : bool, optional
            Print intermediate/debugging information if true, default=False

        Returns
        -------
        list
            [K_0, K_1, ..., K_{N-1}]
        '''
        A, W, H, Q = np.mat(self.kf.A), np.mat(self.kf.W), np.mat(self.kf.H), np.mat(self.kf.Q)
        P = np.mat( np.zeros(A.shape) )
        K = [None]*N
        
        ss_idx = None # index at which K is steady-state (within tol)
        for n in range(N):
            if not ss_idx == None and n > ss_idx:
                K[n] = K[ss_idx]
            else:
                P = A*P*A.T + W 
                K[n] = (P*H.T)*linalg.pinv(H*P*H.T + Q);
                P -= K[n]*H*P;
                if n > 0 and np.linalg.norm(K[n] - K[n-1]) < tol: 
                    ss_idx = n
                    if verbose: 
                        print(("breaking after %d iterations" % n))

        return K, ss_idx

    def get_kf_system_mats(self, T):
        """
        KF system matrices

        x_{t+1} = F_t*x_t + K_t*y_t 

        Parameters
        ----------
        T : int 
            Number of system iterations to calculate (F_t, K_t)

        Returns
        -------
        tuple of lists
            Each element of the tuple is (F_t, K_t) for a given 't'

        """
        F = [None]*T
        K, ss_idx = self.get_kalman_gain_seq(N=T, verbose=False)
        nX = self.kf.A.shape[0]
        I = np.mat(np.eye(nX))
        
        for t in range(T):
            if t > ss_idx: F[t] = F[ss_idx]
            else: F[t] = (I - K[t]*self.kf.H)*self.kf.A
        
        return F, K

    @classmethod
    def MLE_obs_model(self, hidden_state, obs, include_offset=True, drives_obs=None,
        regularizer=None):
        """
        Unconstrained ML estimator of {C, Q} given observations and
        the corresponding hidden states

        Parameters
        ----------
        include_offset : bool, optional, default=True
            A row of all 1's is added as the last row of hidden_state if one is not already present

        Returns
        -------        
        """
        assert hidden_state.shape[1] == obs.shape[1], "different numbers of time samples: %s vs %s" % (str(hidden_state.shape), str(obs.shape))
    
        if isinstance(hidden_state, np.ma.core.MaskedArray):
            mask = ~hidden_state.mask[0,:] # NOTE THE INVERTER 
            inds = np.nonzero([ mask[k]*mask[k+1] for k in range(len(mask)-1)])[0]
    
            X = np.mat(hidden_state[:,mask])
            T = len(np.nonzero(mask)[0])
    
            Y = np.mat(obs[:,mask])
            if include_offset:
                if not np.all(X[-1,:] == 1):
                    X = np.vstack([ X, np.ones([1,T]) ])
        else:
            num_hidden_state, T = hidden_state.shape
            X = np.mat(hidden_state)
            if include_offset:
                if not np.all(X[-1,:] == 1):
                    X = np.vstack([ X, np.ones([1,T]) ])
            Y = np.mat(obs)
    
        n_states = X.shape[0]
        if not drives_obs is None:
            X = X[drives_obs, :]
            
        # ML estimate of C and Q
        if regularizer is None:
            C = np.mat(np.linalg.lstsq(X.T, Y.T)[0].T)
        else:
            x = X.T
            y = Y.T
            XtX_lamb = x.T.dot(x) + regularizer * np.eye(x.shape[1])
            XtY = x.T.dot(y)
            C = np.linalg.solve(XtX_lamb, XtY).T
        Q = np.cov(Y - C*X, bias=1)

        if np.ndim(Q) == 0:
            # if "obs" only has 1 feature, Q might get collapsed to a scalar
            Q = np.mat(Q.reshape(1,1))

        if not drives_obs is None:
            n_obs = C.shape[0]
            C_tmp = np.zeros([n_obs, n_states])
            C_tmp[:,drives_obs] = C
            C = C_tmp
        return (C, Q)
    
    @classmethod 
    def MLE_state_space_model(self, hidden_state, include_offset=True):
        '''
        Train state space model for KF from fully observed hidden state

        Parameters
        ----------
        hidden_state : np.ndarray of shape (N, T)
            N = dimensionality of state vector, T = number of observations
        include_offset : boolean, optional, default=False
            if True, append a "1" to each state vector to add an offset term into the 
            regression

        Returns
        -------        
        A : np.ndarray of shape (N, N)
        W : np.ndarray of shape (N, N)
        '''
        X = hidden_state
        T = hidden_state.shape[1]
        if include_offset:
            X = np.vstack([ X, np.ones([1,T]) ])        
        X1 = X[:,:-1]
        X2 = X[:,1:]
        A = np.linalg.lstsq(X1.T, X2.T)[0].T
        W = np.cov(X2 - np.dot(A, X1), bias=1)
        return A, W

    def set_steady_state_pred_cov(self):
        '''
        Calculate the steady-state prediction covariance and set the current state prediction covariance to the steady-state value
        '''

        A, W, C, Q = np.mat(self.A), np.mat(self.W), np.mat(self.C), np.mat(self.Q)
        D = self.C_xpose_Q_inv_C 
        nS = A.shape[0]
        P = np.mat(np.zeros([nS, nS]))
        I = np.mat(np.eye(nS))

        last_K = np.mat(np.ones(C.T.shape))*np.inf
        K = np.mat(np.ones(C.T.shape))*0

        iter_idx = 0
        for iter_idx in range(40):
            P = A*P*A.T + W
            last_K = K
            KC = P*(I - D*P*(I + D*P).I)*D
            P -= KC*P;

        # TODO fix
        P[0:3, 0:3] = 0
        F, K = self.get_sskf()
        F = (I - KC)*A
        self._init_state(init_state=self.state.mean, init_cov=P)

    def get_K_null(self):
        '''
        $$y_{null} = K_{null} * y_t$$ gives the "null" component of the spike inputs, i.e. $$K_t*y_{null} = 0_{N\times 1}$$
        Parameters
        ----------

        Returns
        -------        
        '''
        F, K = self.get_sskf()
        K = np.mat(K)
        n_neurons = K.shape[1]
        K_null = np.eye(n_neurons) - np.linalg.pinv(K) * K
        return K_null

class KalmanFilterDriftCorrection(KalmanFilter):
    attrs_to_pickle = ['A', 'W', 'C', 'Q', 'C_xpose_Q_inv',
        'C_xpose_Q_inv_C', 'R', 'S', 'T', 'ESS', 'drift_corr','prev_drift_corr']
    noise_threshold = 96.*3.5

    def _init_state(self):
        if hasattr(self, 'prev_drift_corr'):
            self.drift_corr = self.prev_drift_corr.copy()
            print(('prev drift corr', np.mean(self.prev_drift_corr)))
        else:
            self.drift_corr = np.mat(np.zeros(( self.A.shape[0], 1)))
            self.prev_drift_corr = np.mat(np.zeros(( self.A.shape[0], 1)))

        if hasattr(self, 'noise_rej'):
            if self.noise_rej:
                print(('noise rej thresh: ', self.noise_rej_cutoff))
        else:
            self.noise_rej = False
        self.noise_cnt = 0

        super(KalmanFilterDriftCorrection, self)._init_state()
        
    def _forward_infer(self, st, obs_t, Bu=None, u=None, x_target=None, F=None, obs_is_control_independent=True, **kwargs):
        
        if self.noise_rej:
            if np.sum(obs_t) > self.noise_rej_cutoff:
                #print np.sum(obs_t), 'rejecting noise!'
                self.noise_cnt += 1 
                obs_t = np.mat(self.noise_rej_mFR).T
                

        state = super(KalmanFilterDriftCorrection, self)._forward_infer(st, obs_t, Bu=None, u=None, x_target=None, F=None, 
            obs_is_control_independent=True, **kwargs)

        ### Apply Drift Correction ###
        decoded_vel = state.mean.copy()
        state.mean[self.vel_ix] = decoded_vel[self.vel_ix] - self.drift_corr[self.vel_ix]

        ### Update Drift Correcton ###
        self.drift_corr[self.vel_ix] = self.drift_corr[self.vel_ix]*self.drift_rho + decoded_vel[self.vel_ix]*float(1. - self.drift_rho)
        self.prev_drift_corr = self.drift_corr.copy()

        return state

class PCAKalmanFilter(KalmanFilter):
    '''
    A modified KalmanFilter where the Kalman gain is confined to produce outputs in a lower-dimensional linear subspace, i.e. some principal component space
    '''
    def _forward_infer(self, st, obs_t, Bu=None, u=None, target_state=None, obs_is_control_independent=True, **kwargs):
        '''
        See KalmanFilter._forward_infer for docs
        '''
        using_control_input = (Bu is not None) or (u is not None) or (target_state is not None)
        pred_state = self._ssm_pred(st, target_state=target_state, Bu=Bu, u=u)

        C, Q = self.C, self.Q
        P = pred_state.cov

        try:
            M = self.M
            pca_offset = self.pca_offset
        except:
            print("couldn't extract PCA parameters!")
            M = 1
            pca_offset = 0

        K = self._calc_kalman_gain(P)
        I = np.mat(np.eye(self.C.shape[1]))
        D = self.C_xpose_Q_inv_C

        KC = K*C
        F = (I - KC)*self.A

        post_state = pred_state
        if obs_is_control_independent and using_control_input:
            post_state.mean += -KC*self.A*st.mean + M*K*obs_t + pca_offset
        else:
            post_state.mean += -KC*pred_state.mean + M*K*obs_t + pca_offset

        post_state.cov = (I - KC) * P 

        return post_state

    def __getstate__(self):
        '''
        See KalmanFilter.__getstate__ for docs
        '''
        data = super(PCAKalmanFilter, self).__getstate__()
        data['M'] = self.M
        data['pca_offset'] = self.pca_offset
        return data_pickle_

    def __setstate__(self, state):
        '''
        See KalmanFilter.__setstate__ for docs
        '''
        super(PCAKalmanFilter, self).__setstate__(state)
        self.M = state['M']
        self.pca_offset = state['pca_offset']        

class FAKalmanFilter(KalmanFilter):

    def _forward_infer(self, st, obs_t, Bu=None, u=None, target_state=None, obs_is_control_independent=True, **kwargs):
        input_dict = {}
        if hasattr(self, 'FA_kwargs'):

            input_type = self.FA_input + '_input'
        
            input_dict['all_input'] = obs_t.copy()

            dmn = obs_t - self.FA_kwargs['fa_mu']
            shar = (self.FA_kwargs['fa_sharL'] * dmn)
            priv = (dmn - shar)
            main_shar = (self.FA_kwargs['fa_main_shared'] * dmn)
            main_priv = (dmn - main_shar)

            FA = self.FA_kwargs['FA_model']

            inp = obs_t.copy()
            if inp.shape[1] == 1:
                inp = inp.T # want 1 x neurons
            z = FA.transform(dmn.T)
            z = z.T #Transform to fact x 1
            z = z[:self.FA_kwargs['fa_main_shar_n_dim'], :] #only use number in main space

            input_dict['private_input'] = priv + self.FA_kwargs['fa_mu']
            input_dict['shared_input'] = shar + self.FA_kwargs['fa_mu']

            input_dict['private_scaled_input'] = np.multiply(priv, self.FA_kwargs['fa_priv_var_sc']) + self.FA_kwargs['fa_mu']
            input_dict['shared_scaled_input'] = np.multiply(shar, self.FA_kwargs['fa_shar_var_sc']) + self.FA_kwargs['fa_mu']

            input_dict['all_scaled_by_shar_input'] = np.multiply(dmn, self.FA_kwargs['fa_shar_var_sc']) + self.FA_kwargs['fa_mu']

            input_dict['sc_shared+unsc_priv_input'] = input_dict['shared_scaled_input'] + input_dict['private_input'] - self.FA_kwargs['fa_mu']
            input_dict['sc_shared+sc_priv_input'] = input_dict['shared_scaled_input'] + input_dict['private_scaled_input']- self.FA_kwargs['fa_mu']

            input_dict['main_shared_input'] = main_shar + self.FA_kwargs['fa_mu']
            input_dict['main_sc_shared_input'] = np.multiply(main_shar, self.FA_kwargs['fa_main_shared_sc']) + self.FA_kwargs['fa_mu']

            input_dict['main_sc_shar+unsc_priv_input'] = input_dict['main_sc_shared_input'] + input_dict['private_input'] - self.FA_kwargs['fa_mu']
            input_dict['main_sc_shar+sc_priv_input'] = input_dict['main_sc_shared_input'] + input_dict['private_scaled_input'] - self.FA_kwargs['fa_mu']
            input_dict['main_sc_private_input'] = np.multiply(main_priv, self.FA_kwargs['fa_main_private_sc']) + self.FA_kwargs['fa_mu']

            #z = self.FA_kwargs['u_svd'].T*self.FA_kwargs['uut_psi_inv']*dmn
            input_dict['split_input'] = np.vstack((z, main_priv))
            #print input_dict['split_input'].shape
            
            own_pc_trans = np.mat(self.FA_kwargs['own_pc_trans'])*np.mat(dmn)
            input_dict['pca_input'] = own_pc_trans + self.FA_kwargs['fa_mu']

            if input_type in list(input_dict.keys()):
                #print input_type
                obs_t_mod = input_dict[input_type]
            else: 
                print(input_type)
                raise Exception("Error in FA_KF input_type, none of the expected inputs")
        else:
            obs_t_mod = obs_t.copy()

        input_dict['task_input'] = obs_t_mod.copy()


        post_state = super(FAKalmanFilter, self)._forward_infer(st, obs_t_mod, Bu=Bu, u=u, target_state=target_state, 
            obs_is_control_independent=obs_is_control_independent, **kwargs)

        self.FA_input_dict = input_dict

        return post_state

class KFDecoder(bmi.BMI, bmi.Decoder):
    '''
    Wrapper for KalmanFilter specifically for the application of BMI decoding.
    '''
    def __init__(self, *args, **kwargs):
        '''
        Constructor for KFDecoder   
        
        Parameters
        ----------
        *args, **kwargs : see riglib.bmi.bmi.Decoder for arguments
        
        Returns
        -------
        KFDecoder instance
        '''

        super(KFDecoder, self).__init__(*args, **kwargs)
        mFR = kwargs.pop('mFR', 0.)
        sdFR = kwargs.pop('sdFR', 1.)
        self.mFR = mFR
        self.sdFR = sdFR
        self.zeromeanunits = None
        self.zscore = False
        self.kf = self.filt

    def _pickle_init(self):
        super(KFDecoder, self)._pickle_init()
        if not hasattr(self.filt, 'B'):
            self.filt.B = np.mat(np.vstack([np.zeros([3,3]), np.eye(3)*1000*self.binlen, np.zeros(3)]))

        if not hasattr(self.filt, 'F'):
            self.filt.F = np.mat(np.zeros([3,7]))


    def update_params(self, new_params, steady_state=True):
        '''
        Update the decoder parameters if new parameters are available (e.g., by CLDA). See Decoder.update_params
        '''
        super(KFDecoder, self).update_params(new_params)

        # set the KF to the new steady state
        if steady_state:
            self.kf.set_steady_state_pred_cov()

    def __setstate__(self, state):
        """
        Set decoder state after un-pickling. See Decoder.__setstate__, which runs the _pickle_init function at some point during the un-pickling process
        
        Parameters
        ----------
        state : dict
            Variables to set as attributes of the unpickled object.
        
        Returns
        -------
        None
        """
        if 'kf' in state and 'filt' not in state:
            state['filt'] = state['kf']

        super(KFDecoder, self).__setstate__(state)

    def plot_K(self, **kwargs):
        '''
        Plot the Kalman gain weights
        
        Parameters
        ----------
        **kwargs : optional kwargs
            These are passed to the plot function (e.g., which rows to plot)
        
        Returns
        -------
        None
        '''

        F, K = self.kf.get_sskf()
        self.plot_pds(K.T, **kwargs)

    def shuffle(self, shuffle_baselines=False):
        '''
        Shuffle the neural model
        
        Parameters
        ----------
        shuffle_baselines : bool, optional, default = False
            If true, shuffle the estimates of the baseline firing rates in addition to the state-dependent neural tuning parameters.
        
        Returns
        -------
        None (shuffling is done on the current decoder object)        

        '''
        # generate random permutation
        import random
        inds = list(range(self.filt.C.shape[0]))
        random.shuffle(inds)

        # shuffle rows of C, and rows+cols of Q
        C_orig = self.filt.C.copy()
        self.filt.C = self.filt.C[inds, :]
        if not shuffle_baselines:
            self.filt.C[:,-1] = C_orig[:,-1]
        self.filt.Q = self.filt.Q[inds, :]
        self.filt.Q = self.filt.Q[:, inds]

        self.filt.C_xpose_Q_inv = self.filt.C.T * np.linalg.pinv(self.filt.Q.I)

        # RML sufficient statistics (S and T, but not R and ESS)
        # shuffle rows of S, and rows+cols of T
        try:
            self.filt.S = self.filt.S[inds, :]
            self.filt.T = self.filt.T[inds, :]
            self.filt.T = self.filt.T[:, inds]
        except AttributeError:
            # if this decoder never had the RML sufficient statistics
            #   (R, S, T, and ESS) as attributes of self.filt
            pass

    def change_binlen(self, new_binlen, screen_update_rate=60.0):
        '''
        Function to change the binlen of the KFDecoder analytically. 

        Parameters
        ----------
        new_binlen : float
            New bin length of the decoder, in seconds
        screen_update_rate: float, optional, default = 60Hz
            Rate at which the __call__ function will be called
        '''
        bin_gain = new_binlen / self.binlen
        self.binlen = new_binlen

        # Alter bminum, bmicount, # of subbins
        screen_update_period = 1./screen_update_rate
        if self.binlen < screen_update_period:
            self.n_subbins = int(screen_update_period / self.binlen)
            self.bmicount = 0
            if hasattr(self, 'bminum'):
                del self.bminum
        else:
            self.n_subbins = 1
            self.bminum = int(self.binlen / screen_update_period)
            self.bmicount = 0

        # change C matrix
        self.filt.C *= bin_gain
        self.filt.Q *= bin_gain**2
        self.filt.C_xpose_Q_inv *= 1./bin_gain

        # change state space Model
        # TODO generalize this beyond endpoint
        from . import state_space_models
        A, W = self.ssm.get_ssm_matrices(update_rate=new_binlen)
        self.filt.A = A
        self.filt.W = W

    def conv_to_steady_state(self):
        '''
        Create an SSKFDecoder object based on KalmanFilter parameters in this KFDecoder object
        '''
        from . import sskfdecoder
        self.filt = sskfdecoder.SteadyStateKalmanFilter(A=self.filt.A, W=self.filt.W, C=self.filt.C, Q=self.filt.Q) 

    def subselect_units(self, units):
        '''
        Prune units from the KFDecoder, e.g., due to loss of recordings for a particular cell

        Parameters
        units : string or np.ndarray of shape (N,2)
            The units which should be KEPT in the decoder

        Returns 
        -------
        KFDecoder 
            New KFDecoder object using only a subset of the cells of the original KFDecoder
        '''
        # Parse units into list of indices to keep
        inds_to_keep = self._proc_units(units, 'keep')
        dec_new = self._return_proc_units_decoder(inds_to_keep)
        return dec_new
        #self._save_new_dec(dec_new, '_subset')

def project_Q(C_v, Q_hat):
    """ 
    Deprecated! See clda.KFRML_IVC
    """
    print("projecting!")
    from scipy.optimize import fmin_bfgs, fmin_ncg

    C_v = np.mat(C_v)
    Q_hat = np.mat(Q_hat)
    Q_hat_inv = Q_hat.I

    c_1 = C_v[:,0]
    c_2 = C_v[:,1]
    A_1 = c_1*c_1.T - c_2*c_2.T
    A_2 = c_2*c_1.T
    A_3 = c_1*c_2.T
    A = [A_1, A_2, A_3]
    if 1:
        U = np.hstack([c_1 - c_2, c_2, c_1])
        V = np.vstack([(c_1 + c_2).T, c_1.T, c_2.T])
        C_inv_fn = lambda nu: np.mat(np.diag([1./nu[0], 1./(nu[0] + nu[1]), 1./(nu[2] - nu[0]) ]))
        C_fn = lambda nu: np.mat(np.diag([nu[0], (nu[0] + nu[1]), (nu[2] - nu[0]) ]))
        nu_0 = np.zeros(3)
        c_scalars = np.ones(3)
    else:
        u_1, s_1, v_1 = np.linalg.svd(A_1)
        c_scalars = np.hstack([s_1[0:2], 1, 1])
        U = np.hstack([u_1[:,0:2], c_2, c_1])
        V = np.vstack([v_1[0:2, :], c_1.T, c_2.T])
        C_fn = lambda nu: np.mat(np.diag(nu * c_scalars))
        nu_0 = np.zeros(4)

    def cost_fn_gen(nu, return_type='cost'):
        C = C_fn(nu)
        S_star_inv = Q_hat + U*C_fn(nu)*V
        #if return_type == 'cost':
        #    print C_v.T * S_star_inv * C_v
    
        if np.any(np.diag(C) == 0):
            S_star = S_star_inv.I
        else:
            C_inv = C.I
            S_star = Q_hat_inv - Q_hat_inv * U * (C_inv + V*Q_hat_inv*U).I*V * Q_hat_inv;
        
        # log-determinant using LU decomposition, required if Q is large, i.e. lots of simultaneous observations
        cost = -np.log(np.linalg.det(S_star_inv))
        #cost = -np.prod(np.linalg.slogdet(S_star_inv))
        
        # TODO gradient dimension needs to be the same as nu
        #grad = -np.array([np.trace(S_star*U[:,0] * c_scalars[0] * V[0,:]) for k in range(len(nu))])
        #grad = -1e-4*np.array([np.trace(S_star*A[0]), np.trace(S_star*A[1]), np.trace(S_star*A[2])])
        #print c_2.T*S_star*c_2
        grad = -1e-4*np.array(np.hstack([c_1.T*S_star*c_1 - c_2.T*S_star*c_2, c_1.T*S_star*c_2, c_2.T*S_star*c_1])).ravel()
        S = S_star
        hess = np.mat([[np.trace(S*A_1*S*A_1), np.trace(S*A_2*S*A_1), np.trace(S*A_3*S*A_1)],
                       [np.trace(S*A_1*S*A_2), np.trace(S*A_2*S*A_2), np.trace(S*A_3*S*A_2)],
                       [np.trace(S*A_1*S*A_3), np.trace(S*A_2*S*A_3), np.trace(S*A_3*S*A_3)]])
    
        #grad = hess*np.mat(grad.reshape(-1,1))
        #log = logging.getLogger()
        #print "nu = %s, cost = %g, grad=%s" % (nu, cost, grad)
        #log.warning("nu = %s, cost = %g, grad=%s" % (nu, cost, grad))
    
        if return_type == 'cost':
            return cost
        elif return_type == 'grad':
            return grad
        elif return_type == 'hess':
            return hess
        elif return_type == 'opt_val':
            return S_star
        else:
            raise ValueError("Cost function doesn't know how to return this: %s" % return_type)

    cost_fn = lambda nu: cost_fn_gen(nu, return_type = 'cost')
    grad    = lambda nu: cost_fn_gen(nu, return_type = 'grad')
    hess    = lambda nu: cost_fn_gen(nu, return_type = 'hess')
    arg_opt = lambda nu: cost_fn_gen(nu, return_type = 'opt_val')

    # Call optimization routine
    #v_star = fmin_ncg(cost_fn, nu_0, fprime=grad, fhess=hess, maxiter=10000)
    #print v_star
    #v_star = fmin_bfgs(cost_fn, nu_0, maxiter=10000, gtol=1e-15)
    v_star = fmin_bfgs(cost_fn, nu_0, fprime=grad, maxiter=10000, gtol=1e-15)
    print(v_star)

    Q_inv = arg_opt(v_star)
    Q = Q_inv.I
    Q = Q_hat + U * C_fn(v_star) * V

    # TODO print out (log) a more useful measure of success
    #print C_v.T * Q_inv * C_v
    #print C_v.T * Q.I * C_v
    #print v_star
    return Q







'''
Features for the Optitrack motiontracker
'''
from riglib.experiment import traits
from riglib.optitrack_client import optitrack
from datetime import datetime
import time
import numpy as np
import os
from config.rig_defaults import optitrack as defaults
from riglib.stereo_opengl.primitives import Cylinder, Sphere

########################################################################################################
# Optitrack datasources
########################################################################################################
class Optitrack(traits.HasTraits):
    '''
    Enable reading of raw motiontracker data from Optitrack system
    Requires the natnet library from https://github.com/leoscholl/python_natnet
    To be used as a feature with the ManualControl task for the time being. However,
    ideally this would be implemented as a decoder :)
    '''

    optitrack_feature = traits.OptionsList(("rigid body", "skeleton", "marker"))
    smooth_features = traits.Int(1, desc="How many features to average")
    scale = traits.Float(defaults['scale'], desc="Control scale factor")
    offset = traits.Array(value=defaults['offset'], desc="Control offset")
    optitrack_ip = defaults['address']
    optitrack_save_path = defaults['save_path']
    optitrack_sync_dch = defaults['sync_dch']

    hidden_traits = ['optitrack_feature', 'smooth_features']

    def init(self):
        '''
        Secondary init function. See riglib.experiment.Experiment.init()
        Prior to starting the task, this 'init' sets up the DataSource for interacting with the 
        motion tracker system and registers the source with the SinkRegister so that the data gets saved to file as it is collected.
        '''

        # Start the natnet client and recording
        import natnet
        now = datetime.now()
        session = "OptiTrack/Session " + now.strftime("%Y-%m-%d")
        take = now.strftime("Take %Y-%m-%d %H:%M:%S")
        logger = Logger(take)
        try:
            client = natnet.Client.connect(server=self.optitrack_ip, logger=logger, timeout=1)
            if self.saveid is not None:
                take += " (%d)" % self.saveid
                client.set_session(os.path.join(self.optitrack_save_path, session))
                client.set_take(take)
                self.filename = os.path.join(session, take + '.tak')
                client._send_command_and_wait("LiveMode")
                time.sleep(0.1)
                if client.start_recording():
                    self.optitrack_status = 'recording'
            else:
                self.optitrack_status = 'streaming'
        except natnet.DiscoveryError:
            self.optitrack_status = 'Optitrack couldn\'t be started, make sure Motive is open!'
            client = optitrack.SimulatedClient()
        self.client = client

        # Create a source to buffer the motion tracking data
        from riglib import source
        self.motiondata = source.DataSource(optitrack.make(optitrack.System, self.client, self.optitrack_feature, 1))

        # Save to the sink
        from riglib import sink
        sink_manager = sink.SinkManager.get_instance()
        sink_manager.register(self.motiondata)
        super().init()

    def run(self):
        '''
        Code to execute immediately prior to the beginning of the task FSM executing, or after the FSM has finished running. 
        See riglib.experiment.Experiment.run(). This 'run' method starts the motiondata source and stops it after the FSM has finished running
        '''
        if not self.optitrack_status in ['recording', 'streaming']:
            import io
            self.terminated_in_error = True
            self.termination_err = io.StringIO()
            self.termination_err.write(self.optitrack_status)
            self.termination_err.seek(0)
            self.state = None
            super().run()
        else:
            self.motiondata.start()
            try:
                super().run()
            finally:
                print("Stopping optitrack")
                self.client.stop_recording()
                self.motiondata.stop()

    def _start_None(self):
        '''
        Code to run before the 'None' state starts (i.e., the task stops)
        '''
        #self.client.stop_recording()
        self.motiondata.stop()
        super()._start_None()

    def join(self):
        '''
        See riglib.experiment.Experiment.join(). Re-join the motiondata source process before cleaning up the experiment thread
        '''
        if self.optitrack_status in ['recording', 'streaming']:
            print("Joining optitrack datasource")
            self.motiondata.join()
        super().join()

    def cleanup(self, database, saveid, **kwargs):
        '''
        Save the optitrack recorded file into the database
        '''
        super_result = super().cleanup(database, saveid, **kwargs)
        print("Saving optitrack file to database...")
        try:
            database.save_data(self.filename, "optitrack", saveid, False, False) # Make sure you actually have an "optitrack" system added!
        except Exception as e:
            print(e)
            return False
        print("...done.")
        return super_result

    def _get_manual_position(self):
        ''' Overridden method to get input coordinates based on motion data'''

        # Get data from optitrack datasource
        data = self.motiondata.get() # List of (list of features)
        if len(data) == 0: # Data is not being streamed
            return
        recent = data[-self.smooth_features:] # How many recent coordinates to average
        averaged = np.nanmean(recent, axis=0) # List of averaged features
        if np.isnan(averaged).any(): # No usable coords
            return
        return averaged*100 # convert meters to centimeters

class OptitrackSimulate(Optitrack):
    '''
    Fake optitrack data for testing
    '''

    def init(self):
        '''
        Secondary init function. See riglib.experiment.Experiment.init()
        Prior to starting the task, this 'init' sets up the DataSource for interacting with the 
        motion tracker system and registers the source with the SinkRegister so that the data gets saved to file as it is collected.
        '''

        # Start the fake natnet client
        self.client = optitrack.SimulatedClient()
        self.optitrack_status = 'streaming'

        # Create a source to buffer the motion tracking data
        from riglib import source
        self.motiondata = source.DataSource(optitrack.make(optitrack.System, self.client, self.optitrack_feature, 1))

        # Save to the sink
        from riglib import sink
        sink_manager = sink.SinkManager.get_instance()
        sink_manager.register(self.motiondata)
        super(Optitrack, self).init()

class OptitrackPlayback(Optitrack):
    '''
    Read a csv file back into BMI3D as if it were live data
    '''

    filepath = traits.String("", desc="path to optitrack csv file for playback")
    
    def init(self):
        '''
        Secondary init function. See riglib.experiment.Experiment.init()
        Prior to starting the task, this 'init' sets up the DataSource for interacting with the 
        motion tracker system and registers the source with the SinkRegister so that the data gets saved to file as it is collected.
        '''

        # Start the fake natnet client
        self.client = optitrack.PlaybackClient(self.filepath)
        self.optitrack_status = 'streaming'

        # Create a source to buffer the motion tracking data
        from riglib import source
        self.motiondata = source.DataSource(optitrack.make(optitrack.System, self.client, self.optitrack_feature, 1))

        # Save to the sink
        from riglib import sink
        sink_manager = sink.SinkManager.get_instance()
        sink_manager.register(self.motiondata)
        super(Optitrack, self).init()

class HidePlantOnPause():
    '''
    Makes the cursor hidden when the game is paused.
    '''

    def _cycle(self):
        self.plant_visible = not self.pause
        super()._cycle()

class SpheresToCylinders():
    '''
    Convert spheres to cylinders pointing in and out of the screen up to the bounds.
    '''
    
    def add_model(self, model):
        '''
        Hijack spheres and switch them for cylinders along the y-axis
        '''
        if isinstance(model, Sphere) and model.radius > 1:
            height = self.cursor_bounds[3] - self.cursor_bounds[2]
            tmp_model = Cylinder(height, model.radius, color=model.color)
            print('Switched sphere for cylinder')
            model.verts, model.polys, model.tcoords, model.normals = tmp_model.verts, tmp_model.polys, tmp_model.tcoords, tmp_model.normals
            model.rotate_x(90)
        super().add_model(model)


# Helper class for natnet logging
import logging
log_path = os.path.join(os.path.dirname(__file__), '../log/optitrack.log')

class Logger(object):

    def __init__(self, msg="", log_filename=log_path):
        self.log_filename = log_filename
        self.reset(msg)

    def log_str(self, s, mode="a", newline=True):
        if self.log_filename != '':
            if newline and not s.endswith("\n"):
                s += "\n"
            with open(self.log_filename, mode) as fp:
                fp.write(s)
    
    def _log(self, msg, *args):
        self.log_str(msg % args)

    def reset(self, s="Logger"):
        with open(self.log_filename, "w") as fp:
            fp.write(s + "\n\n")

    debug = _log
    info = _log
    warning = _log
    error = _log
    fatal = _log


import pygame
import os

audio_path = os.path.dirname(__file__)

class AudioPlayer():

    def __init__(self, filename='click.wav'):
        if not pygame.mixer.get_init():
            pygame.mixer.pre_init(44100, -16, 2, 2048)
            pygame.mixer.init()
        self.effect = pygame.mixer.Sound(os.path.join(audio_path, filename))

    def play(self):
        self.effect.play()



import datetime
import time
import json
import socket
import select
import os
import multiprocessing as mp
import threading
import queue
import traceback
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.collections import PatchCollection
from matplotlib.widgets import Button, Slider, Cursor
import aopy
from riglib.ecube import MultiSource, map_channels_for_multisource
from riglib.source import MultiChanDataSource

class OnlineDataWorker(threading.Thread):
    '''
    Worker process to receive data from BMI3D.
    '''

    def __init__(self, socket, result_queue):
        '''
        Start the server on the specified IP address and port number.
        '''
        self.sock = socket
        self._stop_event = threading.Event()
        self.result_queue = result_queue
        super().__init__()

    def run(self):
        while not self._stop_event.is_set():
            ready = select.select([self.sock], [], [], 0.1) # 100ms timeout to check _stop_event
            if ready[0]:
                data = self.sock.recv(4096)
                key, value = data.decode('utf-8').split('%')
                self.result_queue.put((key, [json.loads(v) for v in value.split('#')]))

    def stop(self):
        self._stop_event.set()

class AnalysisWorker(mp.Process):
    '''
    Plots eye, cursor, and target data from experiments that have them. Performs automatic
    calibration of eye data to target locations when the cursor enters the target.
    '''

    def __init__(self, task_params, data_queue, figsize=(8,8), update_rate=60, fps=60):
        self.task_params = task_params
        self._stop_event = mp.Event()
        self.data_queue = data_queue
        self.figsize = figsize
        self.update_rate = update_rate
        self.fps = fps
        super().__init__()

    def init(self):
        '''
        Initialize the worker. 
        '''
        self.cycle_count = 0

    def handle_data(self, key, values):
        '''
        Do something with incoming data. By default just keeps track of the time
        '''
        if key == 'cycle_count':
            self.cycle_count = values[0]        

    def draw(self):
        '''
        Update the figure.
        '''
        if 'fps' in self.task_params:
            self.time_text.set_text(f"t={int(self.cycle_count/self.task_params['fps'])} s")
        else:
            self.time_text.set_text(f"t={int(self.cycle_count)} cycles")

    def cleanup(self):
        '''
        Cleanup tasks after the experiment ends, e.g. saving the figure.
        '''
        pass

    def update(self):
        while True:
            try:
                key, values = self.data_queue.get(timeout=0.) # continue if no data
                self.handle_data(key, values)
            except queue.Empty:
                break

    def run(self):
        print('Starting analysis worker:', self.__class__.__name__)
        
        # Initialize figure
        self.fig = plt.figure(figsize=self.figsize)
        self.ax = self.fig.add_axes([0, 0.2, 1, 0.7])
        experiment_name = self.task_params.get('experiment_name', 'None')
        te_id = self.task_params.get('te_id', 'None')
        self.fig.canvas.manager.set_window_title(f"{experiment_name} ({te_id}) - {self.__class__.__name__}")
        self.time_text = self.ax.text(1., 1.05, '', ha='right', va='center', fontsize=12, transform=self.ax.transAxes)
        self.time_text.set_text('Waiting for data...')
        self.init()
                
        # Pop up the figure
        plt.show(block=False)
        plt.pause(0.1)

        t_update = time.perf_counter()
        while not self._stop_event.is_set():
            if time.perf_counter() - t_update > 1./self.update_rate:
                try:
                    self.update()
                except:
                    traceback.print_exc()
                t_update = time.perf_counter()
            self.draw()
            self.time_text.set_text(f"t={int(self.cycle_count/self.task_params['fps'])}s")
            plt.gcf().canvas.draw_idle()
            plt.gcf().canvas.start_event_loop(1./self.fps) 

        self.cleanup()
        plt.close(self.fig)
        print(self.__class__.__name__, 'stopped')

    def stop(self):
        self._stop_event.set()


class BehaviorAnalysisWorker(AnalysisWorker):
    '''
    Plots eye, cursor, and target data from experiments that have them. Performs automatic
    calibration of eye data to target locations when the cursor enters the target if no
    calibration coefficients are available. 
    '''
   
    def __init__(self, task_params, data_queue, calibration_dir='/var/tmp', buffer_time=10, ylim=1, px_per_cm=51.67, **kwargs):
        super().__init__(task_params, data_queue, **kwargs)
        self.calibration_dir = calibration_dir
        self.buffer_time = buffer_time
        self.ylim = ylim
        self.px_per_cm = px_per_cm

    def init(self):
        super().init()
        self.cursor_pos = np.zeros(2)
        self.eye_pos = np.zeros(2)
        self.target_pos = {}
        self.targets = {}
        self.calibration_data = []
        self.calibration_flag = True
        self.eye_coeff = np.array([[1,0],[1,0]])
        self.eye_coeff_corr = 0.5 # Don't accept anything lower than 0.5 by default
        
        self.eye_diam = np.zeros((int(self.buffer_time*self.task_params['fps']), int(2)))

        # Load previous calibration if it exists
        subject = self.task_params.get('subject_name', 'None')
        self.calibration_filename = f'calib_{subject}_{datetime.date.today()}.pkl'
        filepath = os.path.join(self.calibration_dir, self.calibration_filename)
        if os.path.exists(filepath):
            self.eye_coeff, self.eye_coeff_corr = aopy.data.pkl_read(self.calibration_filename, self.calibration_dir)
            self.calibration_flag = False

        # Turn off automatic calibration if this isn't a MC task
        if 'target_radius' not in self.task_params or self.task_params['target_radius'] >= 10:
            self.calibration_flag = False
        if 'rotation' not in self.task_params or self.task_params['rotation'] != 'yzx':
            self.calibration_flag = False

        # Set up figure
        bounds = self.task_params.get('cursor_bounds', (-10,10,0,0,-10,10))
        self.ax.set_xlim(bounds[0], bounds[1])
        self.ax.set_ylim(bounds[-2], bounds[-1])
        self.ax.set_aspect('equal')
        self.circles = PatchCollection([])
        self.ax.add_collection(self.circles)

        # Set up eye diameter figure  
        self.diam_ax = self.fig.add_axes([0.1, 0.06, 0.8, 0.11])
        self.diam_ax.set_ylim(0, self.ylim)
        self.diam_ax.set_xlim(-self.buffer_time, 0)
        self.diam_ax.set_xlabel('Time (s)')
        self.diam_ax.set_ylabel('Eye Diameter (cm)')
        self.diam_plot = self.diam_ax.plot([], [], 'green')[0]


    def update_eye_calibration(self):
        '''
        Update the eye calibration coefficients using the collected data
        '''
        if self.calibration_flag and len(self.calibration_data) > 3:
            eye_data, cursor_data = zip(*self.calibration_data)
            eye_data = np.array(eye_data)
            cursor_data = np.array(cursor_data)
            slopes, intercepts, correlation_coeff = aopy.analysis.fit_linear_regression(eye_data, cursor_data)
            if np.all(abs(correlation_coeff) > self.eye_coeff_corr):
                self.eye_coeff_corr = np.min(abs(correlation_coeff))
                self.eye_coeff = np.vstack((slopes, intercepts)).T
                print('updated calibration:', self.eye_coeff, self.eye_coeff_corr)
            
    def get_current_pos(self):
        '''
        Get the current cursor, eye, and target positions

        Returns:
            cursor_pos ((2,) tuple): Current cursor position
            eye_pos ((2,) tuple): Current eye position and diameters
            targets (list): List of active targets in (position, radius, color) format
        '''
        calibrated_eye_pos = aopy.data.get_calibrated_eye_data(self.eye_pos[0:2], self.eye_coeff)
        try:
            radius = self.task_params['target_radius']
            color = self.task_params['target_color']
            targets = [(self.target_pos[k], radius, color if v == 1 else 'green') for k, v in self.targets.items() if v]
        except:
            targets = []
        return self.cursor_pos, calibrated_eye_pos, targets

    def handle_data(self, key, values):
        super().handle_data(key, values)
        if key == 'sync_event':
            event_name, event_data = values
            if event_name == 'TARGET_ON':
                self.targets[event_data] = 1
            elif event_name == 'TARGET_OFF':
                self.targets[event_data] = 0
            elif event_name in ['PAUSE', 'TRIAL_END', 'HOLD_PENALTY', 'DELAY_PENALTY', 'TIMEOUT_PENALTY']:
                # Clear targets at the end of the trial
                self.targets = {}
            elif event_name == 'REWARD':
                # Set all active targets to reward
                for target_idx in self.targets.keys():
                    self.targets[target_idx] = 2 if self.targets[target_idx] else 0
            elif event_name == 'CURSOR_ENTER_TARGET' and event_data > 0:
                self.calibration_data.append((self.eye_pos, self.cursor_pos))
                self.update_eye_calibration()
        elif key == 'cursor':
            self.cursor_pos = np.array(values[0])[[0,2]]
        elif key == 'eye_pos':
            self.eye_pos = np.array(values[0])[:2]

            # Update eye diameter
            self.temp = np.array(values[0])[4:]
            self.eye_diam = np.roll(self.eye_diam, -1, axis=0)
            self.eye_diam[-1] = self.temp

        elif key == 'target_location':
            target_idx, target_location = values
            self.target_pos[int(target_idx)] = np.array(target_location)[[0,2]]

    def draw(self):
        super().draw()
        cursor_pos, eye_pos, targets = self.get_current_pos()
        cursor_radius = self.task_params.get('cursor_radius', 0.25)
        print(self.eye_diam[0])

        patches = [
            plt.Circle(cursor_pos, cursor_radius), 
            plt.Circle(eye_pos, self.eye_diam[-1, 0]/self.px_per_cm)
        ] + [plt.Circle(pos, radius) for pos, radius, _ in targets]
        self.circles.set_paths(patches)
        colors = ['b', 'g'] + [c for _, _, c in targets]
        self.circles.set_facecolor(colors)
        self.circles.set_alpha(0.5)

        # Update eye diameter plot
        self.diam_plot.set_data(np.arange(len(self.eye_diam)) * 1/(int(self.task_params['fps'])) - 10, 
                                self.eye_diam[:, 0]/self.px_per_cm)

    def cleanup(self):

        # Save the calibration if it was performed
        filepath = os.path.join(self.calibration_dir, self.calibration_filename)
        if self.calibration_flag and not np.array_equal(self.eye_coeff, np.array([[1,0],[1,0]])):
            aopy.data.pkl_write(self.calibration_filename, (self.eye_coeff, self.eye_coeff_corr), self.calibration_dir)

class ERPAnalysisWorker(AnalysisWorker):
    '''
    Plots ERP data from experiments with an ECoG244 array. Automatically calculates 
    ERPs for flash, movement, or laser events depending on the task.
    '''
    bufferlen = 5 # seconds of data to keep in the buffer

    def __init__(self, task_params, data_queue, update_rate=1, time_before=0.05, 
                 time_after=0.1, figure_dir='/home/aolab/figures', **kwargs):
        super().__init__(task_params, data_queue, update_rate=update_rate, **kwargs)
        self.time_before = time_before
        self.time_after = time_after
        self.figure_dir = figure_dir

    def init(self):
        super().init()

        # Initialize the data source
        self.elec_pos, self.acq_ch, _ = aopy.data.load_chmap('ECoG244')
        self.clock_dch = self.task_params.get('screen_sync_dch', 40)
        self.clock_elapsed = 0
        self.clock_times = np.array([])
        self.trigger_dch = None
        self.trigger_events = []
        self.trigger_cycles = []
        self.trigger_times = []
        self.lfp_downsample = 25

        if 'qwalor_trigger_dch' in self.task_params:
            self.trigger_dch = self.task_params['qwalor_trigger_dch']
            channels = map_channels_for_multisource(headstage_channels=self.acq_ch, 
                                                         digital_channels=[self.clock_dch, self.trigger_dch])
        else:
            self.trigger_events.append('TARGET_ON') # For flash
            channels = map_channels_for_multisource(headstage_channels=self.acq_ch, digital_channels=[self.clock_dch])
        self.ds = MultiChanDataSource(MultiSource, channels=channels, bufferlen=self.bufferlen)
        self.ds.start()
        print('datasource started')

        # Initialize the ERP data
        self.erp = np.zeros((int((self.time_before + self.time_after) * self.ds.source.update_freq/self.lfp_downsample), 
                             len(self.elec_pos), 0), dtype=self.ds.source.dtype)
        
        # Initialize the figure
        self.data_map = np.zeros((16,16))
        self.erp_im = aopy.visualization.plot_spatial_map(self.data_map, self.elec_pos[:,0], self.elec_pos[:,1], 
                                                 cmap='bwr', ax=self.ax)
        self.erp_im.set_clim(-100, 100)
        self.erp_text = self.ax.text(0.75, 1.05, '.', ha='center', va='center', fontsize=12, transform=self.ax.transAxes)

        # Add a slider to control the range
        ax_slider = self.fig.add_axes([0.225, 0.075, 0.6, 0.03])
        self.slider = Slider(
            ax=ax_slider,
            label='Range (uV)',
            valmin=0.1,
            valmax=1000,
            valinit=100
        )
        self.slider.on_changed(lambda val: self.erp_im.set_clim(-val, val))

        # And a button to show/hide channel numbers
        self.labels = []
        elec_pos, acq_ch, _ = aopy.data.load_chmap('ECoG244')
        stim_pos, stim_ch, _ = aopy.data.load_chmap('Opto32')
        for pos, ch in zip(elec_pos, acq_ch):
            self.labels.append(aopy.visualization.annotate_spatial_map(pos, ch, 'c', 12, self.ax))
        for pos, ch in zip(stim_pos, stim_ch):
            self.labels.append(aopy.visualization.annotate_spatial_map(pos, ch, 'm', 12, self.ax))
        def toggle_labels(event):
            for label in self.labels:
                label.set_visible(not label.get_visible())
        toggle_labels(None)
        ax_toggle = self.fig.add_axes([0.1, 0.025, 0.1, 0.025])
        self.toggle = Button(ax_toggle, 'Labels')
        self.toggle.on_clicked(toggle_labels)

    def update(self):
        super().update()
        # Keep track of clock cycles
        clock_data = self.ds.get_new(map_channels_for_multisource(digital_channels=[self.clock_dch]))[0]
        if len(clock_data) == 0:
            return # no new data
        timestamps, edges = aopy.utils.detect_edges(clock_data, self.ds.source.update_freq, rising=True, falling=False)
        self.clock_times = np.concatenate((self.clock_times, timestamps + self.clock_elapsed))
        clock_elapsed_prev = self.clock_elapsed
        clock_elapsed_new = self.clock_elapsed + len(clock_data) / self.ds.source.update_freq

        # Check the trigger if it exists
        if self.trigger_dch:
            trigger_data = self.ds.get_new(map_channels_for_multisource(digital_channels=[self.trigger_dch]))[0]
            timestamps, edges = aopy.utils.detect_edges(trigger_data, self.ds.source.update_freq, rising=True, falling=False)   
            self.trigger_times = np.concatenate((self.trigger_times, timestamps + clock_elapsed_prev)).tolist()

        # Append new ERPs from trigger events
        fs = self.ds.source.update_freq / self.lfp_downsample
        nt = 2./self.update_rate # seconds
        npts = int(nt * self.ds.source.update_freq)
        lfp = self.ds.get(npts, map_channels_for_multisource(headstage_channels=self.acq_ch))
        lfp = np.array(lfp)[:,::self.lfp_downsample].T # reshape and downsample (nt, nch)
        ignored_trigger_cycles = []
        while len(self.trigger_cycles) > 0:
            cycle = self.trigger_cycles.pop()
            if cycle >= len(self.clock_times):
                ignored_trigger_cycles.append(cycle)
                continue
            time = self.clock_times[cycle]
            if time + self.time_after > clock_elapsed_new:
                ignored_trigger_cycles.append(cycle)
                continue
            elif time - self.time_before < clock_elapsed_new - nt:
                print('missed cycle', cycle, 'at time', time)
                continue
            erp = aopy.analysis.calc_erp(lfp, [time - clock_elapsed_new + nt], self.time_before, self.time_after, fs)
            self.erp = np.concatenate((self.erp, erp), axis=2)
        self.trigger_cycles = ignored_trigger_cycles

        # Check for new digital triggers
        ignored_trigger_times = []
        while len(self.trigger_times) > 0:
            time = self.trigger_times.pop()
            if time + self.time_after > clock_elapsed_new:
                ignored_trigger_times.append(time)
                continue
            elif time - self.time_before < clock_elapsed_new - nt:
                print('missed trigger', time)
                continue
            erp = aopy.analysis.calc_erp(lfp, [time - clock_elapsed_new + nt], self.time_before, self.time_after, fs)
            self.erp = np.concatenate((self.erp, erp), axis=2)
        self.trigger_times = ignored_trigger_times

        # Update the total elapsed time
        self.clock_elapsed = clock_elapsed_new

        # Update the ERP
        fs = self.ds.source.update_freq / self.lfp_downsample
        max_erp = aopy.analysis.get_max_erp(self.erp, self.time_before, self.time_after, fs, trial_average=True)
        self.data_map = aopy.visualization.get_data_map(max_erp*1.907348633e-7*1e6, self.elec_pos[:,0], self.elec_pos[:,1])

    def handle_data(self, key, values):
        super().handle_data(key, values)
        if key == 'sync_event':
            event_name, event_data = values
            if event_name in self.trigger_events:
                self.trigger_cycles.append(self.cycle_count)

    def draw(self):
        super().draw()
        self.erp_im.set_data(self.data_map)
        self.erp_text.set_text(f"{self.erp.shape[2]} trials")

    def cleanup(self):
        '''
        Cleanup tasks after the experiment ends, e.g. saving the figure.
        '''
        self.ds.stop()      
        # TO-DO: implenent saving figures
        subject = self.task_params.get('subject_name', 'None')
        te_id = self.task_params.get('te_id', 'None')
        if te_id == 'None':
            return
        date = datetime.date.today()
        filename = f'online_erp_{subject}_{te_id}_{date}.png'
        plt.figure(self.fig)
        aopy.visualization.savefig(self.figure_dir, filename, transparent=False)

class OnlineDataServer(threading.Thread):
    '''
    Interface for accumulating and analyzing BMI3D data in real-time.
    '''

    def __init__(self, host_ip, port=5000):
        '''
        Initialize the server on the specified IP address and port number.

        Args:
            host_ip (str): IP address of the machine running the online analysis
            port (int): Port number for the online analysis server
        '''

        # Initialize socket
        self.sock = socket.socket(
            socket.AF_INET, # Internet
              socket.SOCK_DGRAM) # UDP 
        self.sock.bind((host_ip, port))
        self.sock.setblocking(0)
                
        # Initialize workers
        self.data_worker = None
        self.analysis_workers = []
        try:
            MultiSource.pre_init()
            print('eCube streaming initialized')
        except:
            pass

        # Initialize the server
        self._stop_event = threading.Event()
        self.reset()
        self.is_completed = False
        super().__init__()

    def _stop(self):
        # Stop all the workers
        for worker, _ in self.analysis_workers:
            if worker.is_alive():
                worker.stop()
                worker.join()
        self.analysis_workers = []
        if self.data_worker and self.data_worker.is_alive():
            self.data_worker.stop()
            self.data_worker.join()
        self.data_worker = None

    def reset(self):
        self._stop()

        # Start new workers
        self.result_queue = queue.Queue()
        self.data_worker = OnlineDataWorker(self.sock, self.result_queue)
        self.data_worker.start()
        self.is_running = False
        self.is_completed = True
        self.task_params = {}
        self.state = None

    def init(self):
        '''
        Once the experiment is initialized but before it starts, we spin up the analysis processes
        based on what kind of experiment is running.
        '''
        # Always start with the behavior analysis worker
        print('init in state', self.state)
        data_queue = mp.Queue()
        self.analysis_workers.append((BehaviorAnalysisWorker(self.task_params, data_queue), data_queue))

        # Is there an ECoG array?
        if 'record_headstage' in self.task_params and self.task_params['record_headstage']:
            data_queue = mp.Queue()
            self.analysis_workers.append((ERPAnalysisWorker(self.task_params, data_queue), data_queue))

        # Start all the workers
        for worker, _ in self.analysis_workers:
            worker.start()
        
    def update(self):
        '''
        Get the latest data from the online server
        '''
        try:
            key, values = self.result_queue.get(timeout=0.1)
        except queue.Empty:
            time.sleep(0.1)
            return False
        if key == 'state':
            self.state = values[0]
            if values[0] is None:
                print('Experiment finished')
                self.reset()
                self.is_running = False
                self.is_completed = True
            for _, data_queue in self.analysis_workers:
                data_queue.put((key, values))
        elif key == 'init' and values[0]:
            self.is_running = True
            self.init()
        elif key == 'param':
            name, value = values
            self.task_params[name] = value
        else:
            # Send everything else back onto the queues for the analysis workers
            for _, data_queue in self.analysis_workers:
                data_queue.put((key, values))
        return True

    def run(self):
        '''
        Main loop to run the server
        '''
        while True:
            self.update()
            if self._stop_event.is_set():
                break

        self._stop()
        self.sock.close()
        print('OnlineDataServer shut down.')

    def stop(self):
        self._stop_event.set()


if __name__ == '__main__':
    import signal
    import sys

    # Parse arguments
    # online_analysis [hostname] [port] [display]
    if len(sys.argv) >= 2:
        hostname = sys.argv[1]
    else:
        hostname = 'localhost'
    
    if len(sys.argv) >= 3:
        port = sys.argv[2]
    else:
        port = 5000

    if len(sys.argv) >= 4:
        display = sys.argv[3]
    else:
        display = ':0'

    # Spin up servernode
    # if hostname == '0.0.0.0':
    #     import subprocess
    #     subprocess.Popen('/home/aolab/code/bmi3d/riglib/ecube/servernode-control')

    # Start server
    print(hostname, port, display)
    os.environ['DISPLAY'] = display
    analysis = OnlineDataServer(hostname, port)
    analysis.start()

    # Catch interrupts
    def signal_handler(*args, **kwargs):
        analysis.stop()
        sys.exit(0)
    signal.signal(signal.SIGINT, signal_handler)
    signal.pause()




'''
Basic OpenGL shapes constructed out of triangular meshes
'''

import numpy as np
from numpy import pi
try:
    import os
    os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"    
    import pygame
except:
    import warnings
    warnings.warn('riglib/stereo_opengl_primitives.py: not importing name pygame')
import matplotlib.tri as mtri

from .models import TriMesh

class Plane(TriMesh):
    def __init__(self, width=1, height=1, **kwargs):
        pts = np.array([[0,0,0],
                        [width,0,0],
                        [width,height,0],
                        [0,height,0]])
        polys = [(0,1,3),(1,2,3)]
        tcoords = np.array([[0,0],[1,0],[1,1],[0,1]])
        normals = [(0,0,1)]*4
        super(Plane, self).__init__(pts, np.array(polys), 
                tcoords=tcoords, normals=np.array(normals), **kwargs)

class Cube(TriMesh):
    def __init__(self, side_len=1, side_height=None, segments=36, **kwargs):
        self.side_len = side_len
        if side_height is None:
            side_len_half = side_len/2.
        else:
            side_len_half = side_height # 0.5
        side = np.linspace(-1, 1, int(segments/4), endpoint=True)
        
        unit1 = np.hstack(( side[:,np.newaxis], np.ones((len(side),1)), np.ones((len(side),1)) ))
        unit2 = np.hstack(( np.ones((len(side),1)), side[::-1,np.newaxis], np.ones((len(side),1)) ))
        unit3 = np.hstack(( side[::-1,np.newaxis], -1*np.ones((len(side),1)), np.ones((len(side),1)) ))
        unit4 = np.hstack(( -1*np.ones((len(side),1)), side[:,np.newaxis], np.ones((len(side),1)) ))

        unit = np.vstack((unit1, unit2, unit3, unit4))

        pts = np.vstack([unit*[side_len_half, side_len_half, 0], unit*[side_len_half,side_len_half,side_len]])
        normals = np.vstack([unit*[1,1,0], unit*[1,1,0]])
        #pts = np.vstack([unit*[side_len, 0, 0], unit*[side_len,0,side_height]])
        #normals = np.vstack([unit*[1,1,0], unit*[1,1,0]])
        polys = []
        for i in range(segments-1):
            polys.append((i, i+1, i+segments))
            polys.append((i+segments, i+1, i+1+segments))
        polys.append((segments-1, 0, segments*2-1))
        polys.append((segments*2-1, 0, segments))
        
        tcoord = np.array([np.arange(segments), np.ones(segments)]).T
        n = 1./segments
        tcoord = np.vstack([tcoord*[n,1], tcoord*[n,0]])

        super(Cube, self).__init__(pts, np.array(polys), 
            tcoords=tcoord, normals=normals, **kwargs)

class Pipe(TriMesh):
    '''
    Open pipe standing on the xy-plane with the z-axis as the height dimension
    '''
    def __init__(self, height=1, radius=1, segments=36, **kwargs):
        self.height = height
        self.radius = radius
        theta = np.linspace(0, 2*np.pi, segments, endpoint=False)
        unit = np.array([np.cos(theta), np.sin(theta), np.ones(segments)]).T

        pts = np.vstack([unit*[radius, radius, 0], unit*[radius,radius,height]])
        normals = np.vstack([unit*[1,1,0], unit*[1,1,0]])

        polys = []
        for i in range(segments-1):
            polys.append((i, i+1, i+segments))
            polys.append((i+segments, i+1, i+1+segments))
        polys.append((segments-1, 0, segments*2-1))
        polys.append((segments*2-1, 0, segments))
        
        tcoord = np.array([np.arange(segments), np.ones(segments)]).T
        n = 1./segments
        tcoord = np.vstack([tcoord*[n,1], tcoord*[n,0]])

        super().__init__(pts, np.array(polys), 
            tcoords=tcoord, normals=normals, **kwargs)
        
class Disk(TriMesh):
    def __init__(self, radius=1, segments=36, **kwargs):
        pts = [[0, 0, 0]]  # Center point
        
        # Calculate points around the circumference
        angle_increment = 2 * pi / (segments - 2) # not sure why this works
        for i in range(segments):
            x = radius * np.cos(i * angle_increment)
            y = radius * np.sin(i * angle_increment)
            pts.append([x, y, 0])

        pts = np.array(pts)
        
        # Create polygons
        polys = [(0, i, (i + 1) % segments) for i in range(segments)]

        # Texture coordinates
        tcoords = np.zeros((segments + 1, 2))
        tcoords[1:, 0] = np.cos(np.linspace(0, 2 * np.pi, segments))
        tcoords[1:, 1] = np.sin(np.linspace(0, 2 * np.pi, segments))
        
        normals = [(0, 0, 1)] * segments
        
        super().__init__(pts, np.array(polys), 
                                    tcoords=tcoords, normals=np.array(normals), **kwargs)

class Cylinder(TriMesh):
    '''
    Closed cylinder centered on the xy-plane with the z-axis as the height dimension

    Args:
        TriMesh (_type_): _description_
    '''
    def __init__(self, height=1, radius=1, segments=36, **kwargs):
        self.height = height
        self.radius = radius

        body_mesh = Pipe(height, radius, segments)
        top_mesh = Disk(radius, segments)
        bottom_mesh = Disk(radius, segments)

        # Adjust polygon indices for top and bottom based on total number of vertices
        offset_body = len(body_mesh.verts)
        offset_top = offset_body + len(top_mesh.verts)

        total_pts = np.concatenate([body_mesh.verts + np.array([0, 0, -height/2, 0]),
                                    top_mesh.verts + np.array([0, 0, height/2, 0]), 
                                    bottom_mesh.verts + np.array([0, 0, -height/2, 0])])
        total_normals = np.concatenate([body_mesh.normals, top_mesh.normals, bottom_mesh.normals])
        total_tcoords = np.concatenate([body_mesh.tcoords, top_mesh.tcoords, bottom_mesh.tcoords])

        # Update polygons to account for mesh offsets 
        body_polys = body_mesh.polys + 0  # Keep body polygon indices the same 
        top_polys = top_mesh.polys + offset_body
        bottom_polys = bottom_mesh.polys + offset_top

        total_polys = np.concatenate([body_polys, top_polys, bottom_polys])

        # Create the final mesh with combined data
        super().__init__(total_pts, total_polys, tcoords=total_tcoords, normals=total_normals, **kwargs)

class Cable(TriMesh):
    def __init__(self,radius=.5, trajectory = np.array([np.sin(x) for x in range(60)]), segments=12,**kwargs):
        self.trial_trajectory = trajectory
        self.center_value = [0,0,0]
        self.radius = radius
        self.segments = segments
        self.update(**kwargs)
    
    def update(self, **kwargs):
        theta = np.linspace(0, 2*np.pi, self.segments, endpoint=False)
        unit = np.array([np.ones(self.segments),np.cos(theta) ,np.sin(theta)]).T
        intial = np.array([[0,0,self.trial_trajectory[x]] for x in range(len(self.trial_trajectory))])
        self.pts = (unit*[-30/1.36,self.radius,self.radius])+intial[0]
        for i in range(1,len(intial)):
            self.pts = np.vstack([self.pts, (unit*[(i-30)/3,self.radius,self.radius])+intial[i]])

        self.normals = np.vstack([unit*[1,1,0], unit*[1,1,0]])
        self.polys = []
        for i in range(self.segments-1):
            for j in range(len(intial)-1): 
                self.polys.append((i+j*self.segments, i+1+j*self.segments, i+self.segments+j*self.segments))
                self.polys.append((i+self.segments+j*self.segments, i+1+j*self.segments, i+1+self.segments+j*self.segments))

        tcoord = np.array([np.arange(self.segments), np.ones(self.segments)]).T
        n = 1./self.segments
        self.tcoord = np.vstack([tcoord*[n,1], tcoord*[n,0]])
        super(Cable, self).__init__(self.pts, np.array(self.polys), 
            tcoords=self.tcoord, normals=self.normals, **kwargs)

class Torus(TriMesh):
    '''
    Corrected triangle mesh of a torus. Should work well for 3D rendering with lighting.
    '''

    def __init__(self, major_radius=1, minor_radius=0.5, segments_major=36, segments_minor=18, **kwargs):
        self.major_radius = major_radius
        self.minor_radius = minor_radius
        theta_major = np.linspace(0, 2*np.pi, segments_major, endpoint=False)
        phi_minor = np.linspace(0, 2*np.pi, segments_minor, endpoint=False)

        # Create the points for the torus
        pts = []
        for i in range(segments_major):
            for j in range(segments_minor):
                x = (major_radius + minor_radius * np.cos(phi_minor[j])) * np.cos(theta_major[i])
                y = (major_radius + minor_radius * np.cos(phi_minor[j])) * np.sin(theta_major[i])
                z = minor_radius * np.sin(phi_minor[j])
                pts.append([x, y, z])

        pts = np.array(pts)

        # Calculate the normals for the torus using cross product
        normals = []
        for i in range(segments_major):
            for j in range(segments_minor):
                current = pts[i * segments_minor + j]
                next_i = pts[((i + 1) % segments_major) * segments_minor + j]
                next_j = pts[i * segments_minor + (j + 1) % segments_minor]

                edge1 = next_i - current
                edge2 = next_j - current
                normal = np.cross(edge1, edge2)
                normals.append(normal)

        normals = np.array(normals)
        normals /= np.linalg.norm(normals, axis=1)[:, None]

        # Create the polygons for the torus
        polys = []
        for i in range(segments_major):
            for j in range(segments_minor):
                p1 = i * segments_minor + j
                p2 = ((i + 1) % segments_major) * segments_minor + j
                p3 = ((i + 1) % segments_major) * segments_minor + (j + 1) % segments_minor
                p4 = i * segments_minor + (j + 1) % segments_minor
                polys.append((p1, p2, p3))
                polys.append((p3, p4, p1))

        # Create the texture coordinates for the torus
        tcoord = np.array([[i, j] for i in np.linspace(0, 1, segments_major) for j in np.linspace(0, 1, segments_minor)])

        super(Torus, self).__init__(pts, np.array(polys), tcoords=tcoord, normals=normals, **kwargs)

class Sphere(TriMesh):
    def __init__(self, radius=1, segments=36, **kwargs):
        self.radius = radius
        zvals = radius * np.cos(np.linspace(0, np.pi, num=segments))
        circlevals = np.linspace(0, 2*pi, num=segments, endpoint=False)

        vertices = np.zeros(((len(zvals)-2) * len(circlevals), 3))

        for i, z in enumerate(zvals[1:-1]):
            circlepoints = np.zeros((segments, 3))
            circlepoints[:,2] = z
            r = np.sqrt(radius**2 - z**2)
            circlepoints[:,0] = r*np.sin(circlevals)
            circlepoints[:,1] = r*np.cos(circlevals)
            vertices[segments*i:segments*(i+1),:] = circlepoints
        
        vertices = np.vstack([vertices,(0,0,radius),(0,0,-radius)])
        allpointinds = np.arange(len(vertices))
        
        triangles = np.zeros((segments,3))
        firstcirc = allpointinds[0:segments]
        triangles[0,:] = (allpointinds[-2],firstcirc[0], firstcirc[-1])
        for i in range(segments-1):
            triangles[i+1,:] = (allpointinds[-2], firstcirc[i+1], firstcirc[i])
        
        triangles = list(triangles)
        for i in range(segments-3):
            points1 = allpointinds[i*segments:(i+1)*segments]
            points2 = allpointinds[(i+1)*segments:(i+2)*segments]
            for ind, p in enumerate(points1[:-1]):
                t1 = (p, points1[ind+1], points2[ind+1])
                t2 = (p, points2[ind+1], points2[ind])
                triangles += [t1, t2]
            triangles += [(points1[-1], points1[0], points2[0]), (points1[-1], points2[0], points2[-1])]
        
        bottom = np.zeros((segments,3))
        lastcirc = allpointinds[-segments-2:-2]
        bottom[0,:] = (allpointinds[-1], lastcirc[-1], lastcirc[0]) 
        for i in range(segments-1):
            bottom[i+1,:] = (allpointinds[-1], lastcirc[i], lastcirc[i+1])
        triangles = np.vstack([triangles, bottom])
        
        normals = vertices/radius
        hcoord = np.arctan2(normals[:,1], normals[:,0])
        vcoord = np.arctan2(normals[:,2], np.sqrt(vertices[:,0]**2 + vertices[:,1]**2))
        tcoord = np.array([(hcoord+pi) / (2*pi), (vcoord+pi/2) / pi]).T

        super(Sphere, self).__init__(vertices, np.array(triangles), 
            tcoords=tcoord, normals=normals, **kwargs)


class Cone(TriMesh):
    def __init__(self, height=1, radius1=1, radius2=1, segments=36, **kwargs):
        self.height = height
        self.radius1 = radius1
        self.radius2 = radius2
        self.radius = radius1 # for pretending it's a cylinder..
        theta = np.linspace(0, 2*np.pi, segments, endpoint=False)
        unit = np.array([np.cos(theta), np.sin(theta), np.ones(segments)]).T

        pts = np.vstack([unit*[radius1, radius1, 0], unit*[radius2,radius2,height]])
        normals = np.vstack([unit*[1,1,0], unit*[1,1,0]])

        polys = []
        for i in range(segments-1):
            polys.append((i, i+1, i+segments))
            polys.append((i+segments, i+1, i+1+segments))
        polys.append((segments-1, 0, segments*2-1))
        polys.append((segments*2-1, 0, segments))
        
        tcoord = np.array([np.arange(segments), np.ones(segments)]).T
        n = 1./segments
        tcoord = np.vstack([tcoord*[n,1], tcoord*[n,0]])

        super(Cone, self).__init__(pts, np.array(polys), 
            tcoords=tcoord, normals=normals, **kwargs)


class Chain(object):
    '''
    An open chain of cylinders and cones, e.g. to simulate a stick-figure arm/robot
    '''
    def __init__(self, link_radii, joint_radii, link_lengths, joint_colors, link_colors):
        from .models import Group
        from .xfm import Quaternion
        self.num_joints = num_joints = len(link_lengths)

        self.link_radii = self.make_list(link_radii, num_joints)
        self.joint_radii = self.make_list(joint_radii, num_joints)
        self.link_lengths = self.make_list(link_lengths, num_joints)
        self.joint_colors = self.make_list(joint_colors, num_joints)
        self.link_colors = self.make_list(link_colors, num_joints)        

        self.links = []

        # Create the link graphics
        for i in range(self.num_joints):
            joint = Sphere(radius=self.joint_radii[i], color=self.joint_colors[i])

            # The most distal link gets a tapered cylinder (for purely stylistic reasons)
            if i < self.num_joints - 1:
                link = Cylinder(radius=self.link_radii[i], height=self.link_lengths[i], color=self.link_colors[i])
            else:
                link = Cone(radius1=self.link_radii[-1], radius2=self.link_radii[-1]/2, height=self.link_lengths[-1], color=self.link_colors[-1])
            link_i = Group((link, joint))
            self.links.append(link_i)

        link_offsets = [0] + self.link_lengths[:-1]
        self.link_groups = [None]*self.num_joints
        for i in range(self.num_joints)[::-1]:
            if i == self.num_joints-1:
                self.link_groups[i] = self.links[i]
            else:
                self.link_groups[i] = Group([self.links[i], self.link_groups[i+1]])

            self.link_groups[i].translate(0, 0, link_offsets[i])

    def _update_link_graphics(self, curr_vecs):
        from .models import Group
        from .xfm import Quaternion

        for i in range(self.num_joints):
            # Rotate each joint to the vector specified by the corresponding row in self.curr_vecs
            # Annoyingly, the baseline orientation of the first group is always different from the 
            # more distal attachments, so the rotations have to be found relative to the orientation 
            # established at instantiation time.
            if i == 0:
                baseline_orientation = (0, 0, 1)
            else:
                baseline_orientation = (1, 0, 0)

            # Find the normalized quaternion that represents the desired joint rotation
            self.link_groups[i].xfm.rotate = Quaternion.rotate_vecs(baseline_orientation, curr_vecs[i]).norm()

            # Recompute any cached transformations after the change
            self.link_groups[i]._recache_xfm()

    def translate(self, *args, **kwargs):
        self.link_groups[0].translate(*args, **kwargs)

    @staticmethod
    def make_list(value, num_joints):
        '''
        Helper function to allow joint/link properties of the chain to be specified
        as one value for all joints/links or as separate values for each
        '''
        if isinstance(value, list) and len(value) == num_joints:
            return value
        else:
            return [value] * num_joints

##### 2-D primitives #####

class Shape2D(object):
    '''Abstract base class for shapes that live in the 2-dimension xz-plane
    and are intended only for use with the WindowDispl2D class (not Window).
    '''

    def __init__(self, color, visible=True):
        self.color   = color
        self.visible = visible

    def draw(self, surface, pos2pix_fn):
        '''Draw itself on the given pygame.Surface object using the given
        position-to-pixel_position function.'''

        raise NotImplementedError  # implement in subclasses

    def _recache_xfm(self):
        pass


class Circle(Shape2D):
    def __init__(self, center_pos, radius, *args, **kwargs):
        super(Circle, self).__init__(*args, **kwargs)
        self.center_pos = center_pos
        self.radius     = radius

    def draw(self, surface, pos2pix_fn):
        if self.visible:
            color = tuple([int(255*x) for x in self.color[0:3]])

            pix_pos    = pos2pix_fn(self.center_pos)
            pix_radius = pos2pix_fn([self.radius, 0])[0] - pos2pix_fn([0, 0])[0]
            pygame.draw.circle(surface, color, pix_pos, pix_radius)

        return self.visible  # return True if object was drawn


class Sector(Shape2D):
    def __init__(self, center_pos, radius, ang_range, *args, **kwargs):
        super(Sector, self).__init__(*args, **kwargs)
        self.center_pos = center_pos
        self.radius     = radius
        self.ang_range  = ang_range

    def draw(self, surface, pos2pix_fn):
        if self.visible:
            color = tuple([int(255*x) for x in self.color[0:3]])
            
            arc_angles = np.linspace(self.ang_range[0], self.ang_range[1], 5)
            pts = list(self.center_pos + self.radius*np.c_[np.cos(arc_angles), np.sin(arc_angles)])
            pts.append(self.center_pos)
            
            point_list = list(map(pos2pix_fn, pts))
            pygame.draw.polygon(surface, color, point_list)
        
        return self.visible  # return True if object was drawn


class Line(Shape2D):
    def __init__(self, start_pos, length, width, angle, *args, **kwargs):
        super(Line, self).__init__(*args, **kwargs)
        self.start_pos = start_pos
        self.length    = length
        self.width     = width  # draw a line as thin rectangle
        self.angle     = angle

    def draw(self, surface, pos2pix_fn):
        if self.visible:
            color = tuple([int(255*x) for x in self.color[0:3]])

            # create points and then rotate to correct orientation
            pts = np.array([[          0,  self.width/2], 
                            [          0, -self.width/2], 
                            [self.length, -self.width/2], 
                            [self.length,  self.width/2]])
            rot_mat = np.array([[np.cos(self.angle), -np.sin(self.angle)], 
                                [np.sin(self.angle),  np.cos(self.angle)]])
            pts = np.dot(rot_mat, pts.T).T + self.start_pos
            
            point_list = list(map(pos2pix_fn, pts))
            pygame.draw.polygon(surface, color, point_list)

        return self.visible  # return True if object was drawn



'''
High-level classes for BMI used to tie all th BMI subcomponent systems together
'''
import numpy as np
import traceback
import os
import tables
import datetime
import copy
import pickle

class GaussianState(object):
    '''
    Class representing a multivariate Gaussian. Gaussians are
    commonly used to represent the state
    of the BMI in decoders, including the KF and PPF decoders
    '''
    def __init__(self, mean, cov):
        '''
        Parameters
        mean: np.array of shape (N, 1) or (N,)
            N-dimensional vector representing the mean of the multivariate Gaussian distribution
        cov: np.array of shape (N, N)
            N-dimensional covariance matrix
        '''
        if isinstance(mean, np.matrix):
            assert mean.shape[1] == 1 # column vector
            self.mean = mean
        elif isinstance(mean, (float, int)):
            mean = float(mean)
            if isinstance(cov, float):
                self.mean = mean
            else:
                self.mean = mean * np.mat(np.ones([cov.shape[0], 1]))

        elif isinstance(mean, np.ndarray):
            if np.ndim(mean) == 1:
                mean = mean.reshape(-1,1)
            self.mean = np.mat(mean)
        else:
            raise Exception(str(type(mean)))

        # Covariance
        assert cov.shape[0] == cov.shape[1] # Square matrix
        if isinstance(cov, np.ndarray):
            cov = np.mat(cov)
        self.cov = cov

    def __rmul__(self, other):
        '''
        Gaussian RV multiplication:
        If X ~ N(mu, sigma) and A is a matrix, then A*X ~ N(A*mu, A*sigma*A.T)
        '''
        if isinstance(other, np.matrix):
            mu = other*self.mean
            cov = other*self.cov*other.T
        elif isinstance(other, int) or isinstance(other, np.float64) or isinstance(other, float):
            mu = other*self.mean
            cov = other**2 * self.cov
        elif isinstance(other, np.ndarray):
            # This never actually happens for an array because of how numpy implements array multiplication..
            # (but if it did the following would be logical)
            other = np.mat(other)
            mu = other*self.mean
            cov = other*self.cov*other.T
        else:
            raise ValueError("Unrecognized type: ", type(other))
        return GaussianState(mu, cov)

    def __mul__(self, other):
        '''
        Gaussian RV multiplication:
        If X ~ N(mu, sigma) and A is a matrix, then A*X ~ N(A*mu, A*sigma*A.T)
        '''
        mean = other*self.mean
        if isinstance(other, int) or isinstance(other, np.float64) or isinstance(other, float):
            cov = other**2 * self.cov
        else:
            raise ValueError("Unrecognized type: ", type(other))
        return GaussianState(mean, cov)

    def __add__(self, other):
        '''
        Gaussian RV addition: If X ~ N(mu1, sigma1) and Y ~ N(mu2, sigma2), then
        X + Y ~ N(mu1 + mu2, sigma1 + sigma2). If Y is a scalar, then X + Y ~ N(mu1 Y, sigma1)
        '''
        if isinstance(other, (int, float)) and (other == 0):
            return GaussianState(self.mean, self.cov)
        elif isinstance(other, (int, float)):
            return GaussianState(self.mean + other, self.cov)
        elif isinstance(other, GaussianState):
            return GaussianState(self.mean+other.mean, self.cov+other.cov)
        elif isinstance(other, np.matrix) and other.shape == self.mean.shape:
            return GaussianState(self.mean + other, self.cov)
        else:
            # print other
            raise ValueError("Gaussian state: cannot add type :%s" % type(other))

    def probability(self, x, calc_log_pr=False):
        """ Evaluate multivariate Gaussian probability density of input vector, given the
        mean and covariance of this object """
        assert x.shape == self.mean.shape
        k = self.mean.shape[0]

        log_det_sign, log_det_cov = np.linalg.slogdet(self.cov)
        if log_det_sign != 1.0:
            raise ValueError("Covariance matrix is not positive definite!")

        log_pr = -0.5*(k*np.log(2*np.pi) + log_det_cov + \
            (x - self.mean).T * self.cov.I * (x - self.mean))

        if calc_log_pr:
            return log_pr
        else:
            return np.exp(log_pr)

    def volume(self, boundary):
        """ Calculate volume of ellipsoid x^T * cov^-1 * x """
        from scipy.special import gamma

        cov_det = np.linalg.det(self.cov)
        n = self.mean.shape[0]
        if cov_det == 0:
            cov_eigenvals, _ = np.linalg.eig(self.cov)
            n = np.sum(cov_eigenvals > 1e-5)
            cov_det = np.product(cov_eigenvals[cov_eigenvals > 1e-5])

        n_sphere_vol = np.pi**(float(n)/2)/gamma(float(n)/2 + 1)
        return n_sphere_vol * boundary**(float(n)/2) * np.sqrt(cov_det)

    def distance(self, x, sqrt=False):
        """ Calculate Mahalanobis distance """
        if not hasattr(self, "cov_inv"):
            self.cov_inv = np.linalg.inv(self.cov)

        dist_sq = (x - self.mean).T * self.cov_inv * (x - self.mean)
        dist_sq = dist_sq[0,0]
        if sqrt:
            return np.sqrt(dist_sq)
        else:
            return dist_sq



class GaussianStateHMM(object):
    '''
    General hidden Markov model decoder where the state is represented as a Gaussian random vector
    '''
    model_attrs = []

    # List out the attributes to save at pickle time. Might not want this to be every attribute of the decoder (e.g., no point in saving the state of the BMI at pickle-time)
    attrs_to_pickle = []
    def __init__(self, A, W):
        '''
        Constructor for GaussianStateHMM

        x_{t+1} = A*x_t + w_t; w_t ~ N(0, W)

        Parameters
        ----------
        A: np.mat of shape (N, N)
            State transition matrix
        W: np.mat of shape (N, N)
            Noise covariance
        '''
        self.A = A
        self.W = W

    def get_mean(self):
        '''
        Return just the mean of the Gaussian representing the state estimate as a 1D array
        '''
        return np.array(self.state.mean).ravel()

    def _init_state(self, init_state=None, init_cov=None):
        """
        Initialize the state of the filter with a mean and covariance (uncertainty)

        Parameters
        ----------
        init_state : np.matrix, optional
            Initial estimate of the unknown state. If unspecified, a vector of all 0's
            will be used (except for the offset state, if one exists).
        init_cov : np.matrix, optional
            Uncertainty about the initial state. If unspecified, it is assumed that there
            is no uncertainty (a matrix of all 0's).

        Returns
        -------
        None
        """
        ## Initialize the BMI state, assuming
        nS = self.n_states() # number of state variables
        if init_state is None:
            init_state = np.mat( np.zeros([nS, 1]) )
            if self.include_offset: init_state[-1,0] = 1
        if init_cov is None:
            init_cov = np.mat( np.zeros([nS, nS]) )
        self.state = GaussianState(init_state, init_cov)
        self.init_noise_models()

    def n_states(self):
        return self.A.shape[0]

    def init_noise_models(self):
        '''
        Initialize the process and observation noise models. The state noise should be
        Gaussian (as implied by the name of this class). The observation noise may be
        non-Gaussian depending on the observation model.
        '''
        self.state_noise = GaussianState(0.0, self.W)
        self.obs_noise = GaussianState(0.0, self.Q)

    def _ssm_pred(self, state, u=None, Bu=None, target_state=None, F=None):
        '''
        Prior prediction of the hidden states using for linear directed random walk model
        x_{t+1} = Ax_t + c_t + w_t
            x_t = previous state
            c_t = control input (the "directed" part of the model)
            w_t = process noise (the "random walk" part of the model)

        Parameters
        ----------
        state : GaussianState instance
            State estimate and estimator covariance of current state
        u : np.mat, optional, default=None
            An assistive control input. Requires the filter to have an input matrix attribute, B
        Bu : np.mat of shape (N, 1)
            Assistive control input which is precomputed to already account for the control input matrix
        target_state : np.mat of shape (N, 1)
            Optimal value for x_t (defined by external factors, i.e. the task being performed)
        F : np.mat of shape (B.shape[1], N)
            Feedback control gains. Used to compute u_t = BF(x^* - x_t)


        Returns
        -------
        GaussianState instance
            Represents the mean and estimator covariance of the new state estimate
        '''
        A = self.A

        if not (Bu is None):
            c_t = Bu
        elif not (u is None):
            c_t = self.B * u
        elif not (target_state is None):
            B = self.B
            if F is None:
                F = self.F
            # if not np.all(target_state[:-1,:] == 0):
            #     import pdb; pdb.set_trace()
            A = A - B*F
            c_t = B*F*target_state
        else:
            c_t = 0

        return A * state + c_t + self.state_noise

    def __eq__(self, other):
        '''
        Determine equality of two GaussianStateHMM instances
        '''
        # import train
        return GaussianStateHMM.obj_eq(self, other, self.model_attrs)

    def __sub__(self, other):
        '''
        Subtract the model attributes of two GaussianStateHMM instances. Used to determine approximate equality, i.e., equality modulo floating point error
        '''
        # import train
        return GaussianStateHMM.obj_diff(self, other, self.model_attrs)

    @staticmethod
    def obj_eq(self, other, attrs=[]):
        '''
        Determine if two objects have mattching array attributes

        Parameters
        ----------
        other : object
            If objects are not the same type, False is returned
        attrs : list, optional
            List of attributes to compare for equality. Only attributes that are common to both objects are used.
            The attributes should be np.array or similar as np.array_equal is used to determine equality

        Returns
        -------
        bool
            True value returned indicates equality between objects for the specified attributes
        '''
        if isinstance(other, type(self)):
            attrs_eq = [y for y in [x for x in attrs if x in self.__dict__] if y in other.__dict__]
            equal = [np.array_equal(getattr(self, attr), getattr(other, attr)) for attr in attrs_eq]
            return np.all(equal)
        else:
            return False

    @staticmethod
    def obj_diff(self, other, attrs=[]):
        '''
        Calculate the difference of the two objects w.r.t the specified attributes

        Parameters
        ----------
        other : object
            If objects are not the same type, False is returned
        attrs : list, optional
            List of attributes to compare for equality. Only attributes that are common to both objects are used.
            The attributes should be np.array or similar as np.array_equal is used to determine equality

        Returns
        -------
        np.array
            The difference between each of the specified 'attrs'
        '''
        if isinstance(other, type(self)):
            attrs_eq = [y for y in [x for x in attrs if x in self.__dict__] if y in other.__dict__]
            diff = [getattr(self, attr) - getattr(other, attr) for attr in attrs_eq]
            return np.array(diff)
        else:
            return False

    def __call__(self, obs, **kwargs):
        """
        When the object is called directly, it's a wrapper for the
        1-step forward inference function.
        """
        self.state = self._forward_infer(self.state, obs, **kwargs)
        return self.state.mean

    def _pickle_init(self):
        pass

    def __setstate__(self, state):
        """
        Unpickle decoders by loading all the saved parameters and then running _pickle_init

        Parameters
        ----------
        state : dict
            Provided by the unpickling system

        Returns
        -------
        None
        """
        self.__dict__ = state
        self._pickle_init()

    def __getstate__(self):
        data_to_pickle = dict()
        for attr in self.attrs_to_pickle:
            try:
                data_to_pickle[attr] = getattr(self, attr)
            except:
                print(("GaussianStateHMM: could not pickle attribute %s" % attr))
        return data_to_pickle

    def predict(self, observations, *args, **kwargs):
        if isinstance(observations, list):
            N = len(observations)
        elif isinstance(observations, np.ndarray):
            time_dim = kwargs.pop("time_dim", 1)
            if time_dim not in [0, 1]:
                raise ValueError("Can't interpret observation matrix")
            N = observations.shape[time_dim]
            if time_dim == 1:
                observations = observations.T
        state_seq = []
        data = []
        for k in range(N):
            ret = self._forward_infer(self.state, observations[k], *args, **kwargs)
            if np.iterable(ret):
                state_k, data_k = ret
            else:
                state_k = ret
                data_k = None

            self.state = state_k

            state_seq.append(state_k)
            data.append(data_k)
        return state_seq, data


class MachineOnlyFilter(GaussianStateHMM):
    '''
    A degenerate case of the GaussianStateHMM where the inputs are driven only by the
    input control term and all observations are ignored.
    In other words, W = 0 and x_{t+1} = Ax_t + Bu_t
    '''
    def __init__(self, *args, **kwargs):
        super(MachineOnlyFilter, self).__init__(*args, **kwargs)
        self.include_offset = True

    def init_noise_models(self):
        '''
        see bmi.GaussianStateHMM.init_noise_models for documentation
        '''
        self.state_noise = GaussianState(0.0, self.W)

    def _forward_infer(self, st, obs_t, Bu=None, u=None, target_state=None, **kwargs):
        if Bu is not None:
            return self.A * st + Bu
        else:
            return self.A * st


class RectangularBounder(object):
    """ Hard limit on state values """
    def __init__(self, bounding_box, states_to_bound):
        self.bounding_box = bounding_box
        self.states_to_bound = states_to_bound

    def __call__(self, state_mean, state_names):
        """
        Apply bounds on state vector, if bounding box is specified
        """
        state_mean = state_mean.copy()
        min_bounds, max_bounds = self.bounding_box

        repl_with_min = np.array(state_mean[:,0]).ravel() < min_bounds
        repl_with_max = np.array(state_mean[:,0]).ravel() > max_bounds
        state_mean[repl_with_min, :] = min_bounds[repl_with_min].reshape(-1, 1)
        state_mean[repl_with_max, :] = min_bounds[repl_with_max].reshape(-1, 1)
        return state_mean

class Filter(object):
    ''' All Decoder filters should inherit from this abstract class'''

    model_attrs = []
    attrs_to_pickle = []

    def __init__(self):
        '''
        Constructor for LinearScaleFilter
        '''
        self._init_state()

    def get_mean(self):
        ''' Must return self.state.mean to maintain compatibility'''
        return np.array(self.state.mean).ravel()

    def __call__(self, obs, **kwargs):
        ''' 
        Update the state with the new neural observation
        '''
        pass

    def _pickle_init(self):
        pass

    def _init_state(self):
        self.state.mean = 0


class Decoder(object):
    '''
    All BMI decoders should inherit from this class
    '''
    def __init__(self, filt, units, ssm, binlen=0.1, n_subbins=1, tslice=[-1,-1], call_rate=60.0, **kwargs):
        """
        Parameters
        ----------
        filt : PointProcessFilter or KalmanFilter instance
            Generic inference algorithm that does the actual observation decoding
        units : array-like
            N x 2 array of units, where each row is (chan, unit)
        ssm : state_space_models.StateSpace instance
            The state-space model describes the states tracked by the decoder, whether or not
            they are stochastic/related to the observations, bounds on the state, etc.
        binlen : float, optional, default = 0.1
            Bin-length specified in seconds. Gets rounded to a multiple of 1./60
            to match the update rate of the task
        n_subbins : int, optional, default = 3
            Neural observations are always acquired at the 60Hz screen update rate.
            This parameter explains how many bins to sub-divide the observations
            into. Default of 3 is intended to correspond to ~180Hz / 5.5ms bins
        tslice : array_like, optional, default=[-1, -1]
            start and end times for the neural data used to train, e.g. from the .plx file
        call_rate: float, optional, default = 60 Hz
            Rate in Hz at which the task will run the __call__ function.
        """

        self.filt = filt
        if not filt is None:
            self.filt._init_state()
        self.ssm = ssm

        self.units = np.array(units, dtype=np.int32)
        if self.units.size > 0:
            self.channels = np.unique(self.units[:,0])
        else:
            self.channels = []
        self.binlen = binlen
        self.bounding_box = ssm.bounding_box
        self.states = ssm.state_names

        # The tslice parameter below properly belongs in the database and
        # not in the decoder object because the Decoder object has no record of
        # which plx file it was trained from. This is a leftover from when it
        # was assumed that every decoder would be trained entirely from a plx
        # file (i.e. and not CLDA)
        self.tslice = tslice
        self.states_to_bound = ssm.states_to_bound

        self.drives_neurons = ssm.drives_obs #drives_neurons
        self.n_subbins = n_subbins

        self.bmicount = 0
        self.bminum = int(self.binlen/(1/call_rate))
        self.spike_counts = np.zeros([len(units), 1])

        self.set_call_rate(call_rate)

        self._pickle_init()

    def _pickle_init(self):
        '''
        Functionality common to unpickling a Decoder from file and instantiating a new Decoder.
        A call to this function is the last line in __init__ as well as __setstate__.
        '''
        from . import train

        # If the decoder doesn't have an 'ssm' attribute, then it's an old
        # decoder in which case the ssm is the 2D endpoint SSM
        if not hasattr(self, 'ssm'):
            from . import state_space_models
            self.ssm = state_space_models.StateSpaceEndptVel2D()
            # self.ssm = train.endpt_2D_state_space

        # Assign a default call rate of 60 Hz and initialize the bmicount/bminum attributes
        if hasattr(self, 'call_rate'):
            self.set_call_rate(self.call_rate)
        else:
            self.set_call_rate(60.0)


    def init_zscore(self, mFR_curr, sdFR_curr):
        '''
        Initialize parameters for zcoring observations, if that feature is enabled in the decoder object
        
        Parameters
        ----------
        mFR_curr : np.array of shape (N,)
            Current mean estimates (as opposed to potentially old estimates already stored in the decoder)
        sdFR_curr : np.array of shape (N,)
            Current standard deviation estimates (as opposed to potentially old estimates already stored in the decoder)
        
        Returns
        -------
        None
        '''

        # if interfacing with Kinarm system, may mean and sd will be shape (n, 1)
        self.zeromeanunits, = np.nonzero(mFR_curr == 0) #find any units with a mean FR of zero for this session
        sdFR_curr[self.zeromeanunits] = np.nan # set mean and SD of quiet units to nan to avoid divide by 0 error
        mFR_curr[self.zeromeanunits] = np.nan
        #self.sdFR_ratio = self.sdFR/sdFR_curr
        #self.mFR_diff = mFR_curr-self.mFR
        #self.mFR_curr = mFR_curr
        self.mFR = mFR_curr
        self.sdFR = sdFR_curr
        self.zscore = True

        
    def plot_pds(self, C, ax=None, plot_states=['hand_vx', 'hand_vz'], invert=False, **kwargs):
        '''
        Plot 2D "preferred directions" of features in the Decoder

        Parameters
        ----------
        C: np.array of shape (n_features, n_states)
        ax: matplotlib.pyplot axis, default=None
            axis to plot on. If None specified, a new one is created.
        plot_states: list of strings, default=['hand_vx', 'hand_vz']
            List of decoder states to plot. Only two can be specified currently
        invert: bool, default=False
            If true, flip the signs of the arrows plotted
        kwargs: dict
            Keyword arguments for the low-level matplotlib function
        '''
        import matplotlib.pyplot as plt
        if ax == None:
            plt.figure()
            ax = plt.subplot(111)
            ax.hold(True)

        if C.shape[1] > 2:
            state_inds = [self.states.index(x) for x in plot_states]
            x, z = state_inds
        else:
            x, z = 0, 1
        n_neurons = C.shape[0]
        linestyles = ['-.', '-', '--', ':']
        if invert:
            C = C*-1
        for k in range(n_neurons):
            unit_str = '%d%s' % (self.units[k,0], chr(96 + self.units[k,1]))
            ax.plot([0, C[k, x]], [0, C[k, z]], label=unit_str, linestyle=linestyles[k/7 % len(linestyles)], **kwargs)
        ax.legend(bbox_to_anchor=(1.1, 1.05), prop=dict(size=8))
        try:
            ax.set_xlabel(plot_states[0])
            ax.set_ylabel(plot_states[1])
        except:
            pass
        ax.set_title(self)

    def plot_C(self, **kwargs):
        '''
        Plot the C matrix (see plot_pds docstring), which is used
        by the KFDecoder and the PPFDecoder
        '''
        self.plot_pds(self.filt.C, linewidth=2, **kwargs)

    def update_params(self, new_params, **kwargs):
        '''
        Method for updating the parameters of the decoder

        Parameters
        ----------
        new_params: dict
            Keys are the parameters to be replaced, values are the new value of
            the parameter to replace. In particular, the keys can be dot-separated,
            e.g. to set the attribute 'self.kf.C', the key would be 'kf.C'
        '''
        for key, val in list(new_params.items()):
            attr_list = key.split('.')
            final_attr = attr_list[-1]
            attr_list = attr_list[:-1]
            attr = self
            while len(attr_list) > 0:
                attr = getattr(self, attr_list[0])
                attr_list = attr_list[1:]

            setattr(attr, final_attr, val)

    def bound_state(self):
        """
        Apply bounds on state vector, if bounding box is specified
        """
        if not self.bounding_box is None:
            min_bounds, max_bounds = self.bounding_box
            state = self[self.states_to_bound]
            repl_with_min = state < min_bounds
            state[repl_with_min] = min_bounds[repl_with_min]

            repl_with_max = state > max_bounds
            state[repl_with_max] = max_bounds[repl_with_max]
            self[self.states_to_bound] = state

    def __getitem__(self, idx):
        """
        Get element(s) of the BMI state, indexed by name or number

        Warning: The variable 'q' is a reserved keyword, referring to all of
        the position states. This strange letter choice was made to be consistent
        with the robotics literature, where 'q' refers to the vector of
        generalized joint coordinates.

        Parameters
        ----------
        idx: int or string
            Name of the state, index of the state, or list of indices/names
            of the Decoder state(s) to return
        """
        if isinstance(idx, int) or isinstance(idx, np.int64) or isinstance(idx, np.int32):
            return self.filt.state.mean[idx, 0]
        elif idx == 'q':
            pos_states, = np.nonzero(self.ssm.state_order == 0)
            return np.array([self.__getitem__(k) for k in pos_states])
        elif idx == 'qdot':
            vel_states, = np.nonzero(self.ssm.state_order == 1)
            return np.array([self.__getitem__(k) for k in vel_states])
        elif isinstance(idx, str) or isinstance(idx, str):
            idx = self.states.index(idx)
            return self.filt.state.mean[idx, 0]
        elif np.iterable(idx):
            return np.array([self.__getitem__(k) for k in idx])
        else:
            try:
                return self.filt.state.mean[idx, 0]
            except:
                raise ValueError("Decoder: Improper index type: %s" % type(idx))

    def __setitem__(self, idx, value):
        """
        Set element(s) of the BMI state, indexed by name or number

        Parameters
        ----------
        idx: int or string
            Name of the state, index of the state, or list of indices/names
            of the Decoder state(s) to return
        """
        if isinstance(idx, int) or isinstance(idx, np.int64) or isinstance(idx, np.int32):
            self.filt.state.mean[idx, 0] = value
        elif idx == 'q':
            pos_states, = np.nonzero(self.ssm.state_order == 0)
            self.filt.state.mean[pos_states, 0] = value
        elif idx == 'qdot':
            vel_states, = np.nonzero(self.ssm.state_order == 1)
            self.filt.state.mean[vel_states, 0] = value
        elif isinstance(idx, str) or isinstance(idx, str):
            idx = self.states.index(idx)
            self.filt.state.mean[idx, 0] = value
        elif np.iterable(idx):
            [self.__setitem__(k, val) for k, val in zip(idx, value)]
        else:
            try:
                self.filt.state.mean[idx, 0] = value
            except:
                raise ValueError("Decoder: Improper index type: %" % type(idx))

    def __setstate__(self, state):
        """
        Set decoder state after un-pickling
        """
        if 'db_entry' in state:
            del state['db_entry']
        self.__dict__.update(state)
        self.filt._pickle_init()
        self.filt._init_state()

        if not hasattr(self, 'n_subbins'):
            self.n_subbins = 1

        if not hasattr(self, 'interpolate_using_ssm'):
            self.interpolate_using_ssm = False

        if not hasattr(self, 'bmicount'):
            self.bmicount = 0

        if not hasattr(self, 'n_features'):
            self.n_features = len(self.units)

        self._pickle_init()

    def set_call_rate(self, call_rate):
        '''
        Function for the higher-level task to set the frequency of function calls to __call__

        Parameters
        ----------
        call_rate : float
            1./call_rate should be an integer multiple or divisor of the Decoder's 'binlen'

        Returns
        -------
        None
        '''
        self.call_rate = call_rate
        self.bmicount = 0
        self.bminum = int(self.binlen/(1./self.call_rate))
        self.n_subbins = int(np.ceil(1./self.binlen /self.call_rate))

    def get_state(self, shape=-1):
        '''
        Get the state of the decoder (mean of the Gaussian RV representing the
        state of the BMI)
        '''
        return np.asarray(self.filt.state.mean).reshape(shape)

    def predict(self, neural_obs, assist_level=0.0, weighted_avg_lfc=False, **kwargs):
        """
        Decode the spikes

        Parameters
        ----------
        neural_obs: np.array of shape (N,) or (N, 1)
            One time-point worth of neural features to decode
        assist_level: float
            Weight given to the assist term. This variable name may be a slight misnomer, a more appropriate term might be 'reweight_factor'
        Bu: np.mat of shape (N, 1)
            Assist vector to be added on to the Decoder state. Must be of the same dimension
            as the state vector.
        kwargs: dict
            Mostly for kwargs function call compatibility
        """
        if np.any(neural_obs > 1000):
            print('observations have counts >> 1000 ')

        if np.any(assist_level) > 0 and 'x_assist' not in kwargs:
            raise ValueError("Assist cannot be used if the forcing term is not specified!")

        # re-normalize the variance of the spike observations, if nec
        if hasattr(self, 'zscore') and self.zscore:
            #neural_obs = (np.asarray(neural_obs).ravel() - self.mFR_curr) * self.sdFR_ratio
            neural_obs = (np.asarray(neural_obs).ravel() - self.mFR) * (1./self.sdFR)
            # set the spike count of any unit that now has zero-mean with its original mean
            # This functionally removes it from the decoder.
            neural_obs[self.zeromeanunits] = self.mFR[self.zeromeanunits]

        # re-format as a column matrix
        neural_obs = np.mat(neural_obs.reshape(-1,1))

        x = self.filt.state.mean

        # Run the filter
        self.filt(neural_obs, **kwargs)

        if np.any(assist_level) > 0:
            x_assist = kwargs.pop('x_assist')

            if 'ortho_damp_assist' in kwargs and kwargs['ortho_damp_assist']:
                x_assist[self.drives_neurons,:] /= np.linalg.norm(x_assist[self.drives_neurons,:])
                targ_comp = float(self.filt.state.mean[self.drives_neurons,:].T*x_assist[self.drives_neurons,:])*x_assist[self.drives_neurons,:]
                orth_comp = self.filt.state.mean[self.drives_neurons,:] - targ_comp

                if type(assist_level) is np.ndarray:
                    tmp = np.mat(np.zeros((len(self.filt.state.mean)))).T
                    tmp[:7, :] = self.filt.state.mean[:7, 0]
                    assist_level_ix = kwargs['assist_level_ix']
                    for ia, al in enumerate(assist_level):
                        ix = np.nonzero(assist_level_ix[ia] <= 6)[0]
                        tmp[assist_level_ix[ia][ix]+7, :] = targ_comp[assist_level_ix[ia][ix]] + (1 - al)*orth_comp[assist_level_ix[ia][ix]]
                    self.filt.state.mean = tmp

                else:
                    # High assist damps orthogonal component a lot
                    self.filt.state.mean[self.drives_neurons,:] = targ_comp + (1 - assist_level)*orth_comp

            elif type(assist_level) is np.ndarray:
                tmp = np.zeros((len(self.filt.state.mean)))
                assist_level_ix = kwargs['assist_level_ix']
                for ia, al in enumerate(assist_level):
                    tmp[assist_level_ix[ia]] = (1-al)*self.filt.state.mean[assist_level_ix[ia]] + al*x_assist[assist_level_ix[ia]]
                self.filt.state.mean = np.mat(tmp).T

            else:
                self.filt.state.mean = (1-assist_level)*self.filt.state.mean + assist_level * x_assist.reshape(-1,1)

        # Bound cursor, if any hard bounds for states are applied
        if hasattr(self, 'bounder'):
            self.filt.state.mean = self.bounder(self.filt.state.mean, self.states)

        state = self.filt.get_mean()
        return state

    def decode(self, neural_obs, **kwargs):
        '''
        Decode multiple observations sequentially.

        Parameters
        ----------
        neural_obs: np.array of shape (# features, # observations)
            Independent neural observations are columns of the data
            matrix and are decoded sequentially
        kwargs: dict
            Container for special keyword-arguments for the specific decoding
            algorithm's 'predict'.
        '''
        output = []
        n_obs = neural_obs.shape[1]
        for k in range(n_obs):
            self.predict(neural_obs[:,k], **kwargs)
            output.append(self.filt.get_mean())
        return np.vstack(output)

    def __str__(self):
        if hasattr(self, 'db_entry'):
            return self.db_entry.name
        else:
            return super(Decoder, self).__str__()

    @property
    def n_states(self):
        '''
        Return the number of states represented in the Decoder
        '''
        return len(self.states)

    @property
    def n_units(self):
        '''
        Return the number of units used in the decoder. Not sure what this
        does for LFP decoders, i.e. decoders which extract multiple features from
        a single channel.
        '''
        return len(self.units)

    def __call__(self, obs_t, **kwargs):
        '''
        Wrapper for the 'predict' method

        Parameters
        ----------
        obs_t: np.array of shape (# features, # subbins)
            Neural observation vector. If the decoding_rate of the Decoder is
            greater than the control rate of the plant (e.g. 60 Hz )
        kwargs: dictionary
            Algorithm-specific arguments to be given to the Decoder.predict method
        '''

        self.predict(obs_t, **kwargs)
        return self.filt.get_mean().reshape(-1,1)

    def save(self, filename=''):
        '''
        Pickle the Decoder object to a file

        Parameters
        ----------
        filename: string, optional
            Filename to pickle the decoder to. If unspecified, a temporary file will be created.

        Returns
        -------
        filename: string
            filename of pickled Decoder object
        '''
        if filename != '':
            f = open(filename, 'wb')
            pickle.dump(self, f)
            f.close()
            return filename
        else:
            import tempfile
            tf2 = tempfile.NamedTemporaryFile(delete=False)
            pickle.dump(self, tf2)
            tf2.flush()
            return tf2.name

    def save_attrs(self, hdf_filename, table_name='task'):
        '''
        Save the attributes of the Decoder to the attributes of the specified HDF table

        Parameters
        ----------
        hdf_filename: string
            HDF filename to write data to
        table_name: string, default='task'
            Specify the table within the HDF file to set attributes in.
        '''
        h5file = tables.openFile(hdf_filename, mode='a')
        table = getattr(h5file.root, table_name)
        for attr in self.filt.model_attrs:
            table.attrs[attr] = np.array(getattr(self.filt, attr))
        h5file.close()

    @property
    def state_shape_rt(self):
        '''
        Create attribute to access the shape of the accumulating spike counts feature.
        '''
        return (self.n_states, self.n_subbins)


class BMISystem(object):
    '''
    This class encapsulates all of the BMI decoding computations, including assist and CLDA
    '''
    def __init__(self, decoder, learner, updater, feature_accumulator):
        '''
        Instantiate the BMISystem

        Parameters
        ----------
        decoder : bmi.Decoder instance
            The decoder maps spike counts into the "state" of the prosthesis
        learner : clda.Learner instance
            The learner estimates the "intended" prosthesis state from task goals.
        updater : clda.Updater instance
            The updater remaps the decoder parameters to better match sets of
            observed spike counts and intended kinematics (from the learner)
        feature_accumulator : accumulator.FeatureAccumulator instance
            Combines features across time if necesary to perform rate matching
            between the task rate and the decoder rate.

        Returns
        -------
        BMISystem instance
        '''
        self.decoder = decoder
        self.learner = learner
        self.updater = updater
        self.feature_accumulator = feature_accumulator
        self.param_hist = []

        self.has_updater = not (self.updater is None)
        if self.has_updater:
            self.updater.init(self.decoder)

    def __call__(self, neural_obs, target_state, task_state, learn_flag=False, **kwargs):
        '''
        Main function for all BMI functions, including running the decoder, adapting the decoder
        and incorporating assistive control inputs

        Parameters
        ----------
        neural_obs : np.ndarray,
            The shape of neural_obs should be [n_units, n_obs]. If multiple observations are given, then
            the decoder will run multiple times before returning.
        target_state : np.ndarray
            The assumed state that the subject is trying to drive the BMI toward, e.g. based on the
            objective of the task
        task_state : string
            State of the task. Used by CLDA so that assist is only applied during certain states,
            e.g. in some tasks, the target will be ambiguous during penalty states so CLDA should
            ignore data during those epochs.
        learn_flag : bool, optional, default=True
            Boolean specifying whether the decoder should update based on intention estimates
        **kwargs : dict
            Instance-specific arguments, e.g. RML/SmoothBatch require a 'half_life' parameter
            that is not required of other CLDA methods.

        Returns
        -------
        decoded_states : np.ndarray
            Columns of the array are vectors representing the decoder output as each of the
            observations are decoded.
        update_flag : boolean
            Boolean to indicate whether the parameters of the Decoder have changed based on the
            current function call
        '''
        n_units, n_obs = neural_obs.shape
        # If the target is specified as a 1D position, tile to match
        # the number of dimensions as the neural features
        if np.ndim(target_state) == 1 or (target_state.shape[1] == 1 and n_obs > 1):
            target_state = np.tile(target_state, [1, n_obs])


        decoded_states = np.zeros([self.decoder.n_states, n_obs])
        update_flag = False

        for k in range(n_obs):
            neural_obs_k = neural_obs[:,k].reshape(-1,1)
            target_state_k = target_state[:,k]

            # NOTE: the conditional below is *only* for compatibility with older Carmena
            # lab data collected using a different MATLAB-based system. In all python cases,
            # the task_state should never contain NaN values.
            if np.any(np.isnan(target_state_k)): task_state = 'no_target'

            #################################
            ## Decode the current observation
            #################################
            decodable_obs, decode = self.feature_accumulator(neural_obs_k)
            if decode: # if a new decodable observation is available from the feature accumulator
                prev_state = self.decoder.get_state()

                self.decoder(decodable_obs, **kwargs)

                # Determine whether the current state or previous state should be given to the learner
                if self.learner.input_state_index == 0:
                    learner_state = self.decoder.get_state()
                elif self.learner.input_state_index == -1:
                    learner_state = prev_state
                else:
                    print(("Not implemented yet: %d" % self.learner.input_state_index))
                    learner_state = prev_state

                if learn_flag:
                    self.learner(decodable_obs.copy(), learner_state, target_state_k, self.decoder.get_state(), task_state, state_order=self.decoder.ssm.state_order)

            decoded_states[:,k] = self.decoder.get_state()

            ############################
            ## Update decoder parameters
            ############################
            if self.learner.is_ready():
                batch_data = self.learner.get_batch()
                batch_data['decoder'] = self.decoder
                kwargs.update(batch_data)
                self.updater(**kwargs)
                self.learner.disable()

            new_params = None # by default, no new parameters are available
            if self.has_updater:
                new_params = copy.deepcopy(self.updater.get_result())

            # Update the decoder if new parameters are available
            if not (new_params is None):
                self.decoder.update_params(new_params, **self.updater.update_kwargs)
                new_params['intended_kin'] = batch_data['intended_kin']
                new_params['spike_counts_batch'] = batch_data['spike_counts']
                new_params['target_position'] = target_state_k

                self.learner.enable()
                update_flag = True

                # Save new parameters to parameter history
                self.param_hist.append(new_params)
        return decoded_states, update_flag


class BMILoop(object):
    '''
    Container class/interface definition for BMI tasks. Intended to be used with multiple inheritance structure paired with riglib.experiment classes
    '''
    static_states = [] # states in which the decoder is not run
    decoder_sequence = ''

    def init(self):
        '''
        Secondary init function. Finishes initializing the task after all the
        constructors have run and all the requried attributes have been declared
        for the task to operate.
        ''' 
        # Make sure the plant is up to date
        self.plant.set_endpoint_pos(np.array(self.starting_pos))

        # Initialize the decoder
        self.load_decoder()
        self.init_decoder_state()
        if hasattr(self.decoder, 'adapting_state_inds'):
            print('Decoder has adapting state inds')

        if hasattr(self.decoder, 'adapting_neural_inds'):
            print('Decoder has adapting neural inds')

        # Declare data attributes to be stored in the sinks every iteration of the FSM
        self.add_dtype('loop_time', 'f8', (1,))
        self.add_dtype('decoder_state', 'f8', (self.decoder.n_states, 1))
        self.add_dtype('internal_decoder_state', 'f8', self.decoder.state_shape_rt)
        self.add_dtype('target_state', 'f8', self.decoder.state_shape_rt)
        self.add_dtype('update_bmi', 'f8', (1,))

        # Construct the sub-pieces of the BMI system
        self.create_assister()
        self.create_feature_extractor()
        self.create_feature_accumulator()
        self.create_goal_calculator()
        self.create_learner()
        self.create_updater()
        self.create_bmi_system()

        super(BMILoop, self).init()

    def create_bmi_system(self):
        self.bmi_system = BMISystem(self.decoder, self.learner, self.updater, self.feature_accumulator)

    def load_decoder(self):
        '''
        Shell function. In tasks launched from the GUI with the BMI feature
        enabled, the decoder attribute is automatically added to the task. This
        is for simulation purposes only (or if you want to make a version that
        launches from the command line)
        '''
        pass

    def init_decoder_state(self):
        '''
        Initialize the state of the decoder to match the initial state of the plant
        '''
        self.decoder.filt._init_state()
        try:
            self.decoder['q'] = self.plant.get_intrinsic_coordinates()
        except:
            print((self.plant.get_intrinsic_coordinates()))
            print((self.decoder['q']))
            raise Exception("Error initializing decoder state")
        self.init_decoder_mean = self.decoder.filt.state.mean.copy()

        self.decoder.set_call_rate(1./self.update_rate)

    def create_assister(self):
        '''
        The 'assister' is a callable object which, for the specific plant being controlled,
        will drive the plant toward the specified target state of the task.
        '''
        self.assister = None

    def create_feature_accumulator(self):
        '''
        Instantiate the feature accumulator used to implement rate matching between the Decoder and the task,
        e.g. using a 10 Hz KFDecoder in a 60 Hz task
        '''
        from . import accumulator
        feature_shape = [self.decoder.n_features, 1]
        feature_dtype = np.float64
        acc_len = int(self.decoder.binlen / self.update_rate)
        acc_len = max(1, acc_len)
        if self.extractor.feature_type in ['lfp_power', 'emg_amplitude']:
            self.feature_accumulator = accumulator.NullAccumulator(acc_len)
        else:
            self.feature_accumulator = accumulator.RectWindowSpikeRateEstimator(acc_len, feature_shape, feature_dtype)

    def create_goal_calculator(self):
        '''
        The 'goal_calculator' is a callable object which will define the optimal state for the Decoder
        to be in for this particular task. This object is necessary for CLDA (to estimate the "error" of the decoder
        in order to adapt it) and for any assistive control (for the 'machine' controller to determine where to
        drive the plant
        '''
        self.goal_calculator = None

    def create_feature_extractor(self):
        '''
        Create the feature extractor object. The feature extractor takes raw neural data from the streaming processor
        (e.g., spike timestamps) and outputs a decodable observation vector (e.g., counts of spikes in last 100ms from each unit)
        '''
        from . import extractor
        if hasattr(self.decoder, 'extractor_cls') and hasattr(self.decoder, 'extractor_kwargs'):
            self.extractor = self.decoder.extractor_cls(self.neurondata, **self.decoder.extractor_kwargs)
        else:
            # if using an older decoder that doesn't have extractor_cls and
            # extractor_kwargs as attributes, then create a BinnedSpikeCountsExtractor by default
            self.extractor = extractor.BinnedSpikeCountsExtractor(self.neurondata,
                n_subbins=self.decoder.n_subbins, units=self.decoder.units)

        self._add_feature_extractor_dtype()

    def _add_feature_extractor_dtype(self):
        '''
        Helper function to add the datatype of the extractor output to be saved in the HDF file. Uses a separate function
        so that simulations can overwrite.
        '''
        if isinstance(self.extractor.feature_dtype, tuple): # Feature extractor only returns 1 type
            self.add_dtype(*self.extractor.feature_dtype)
        else:
            for x in self.extractor.feature_dtype: # Feature extractor returns multiple named fields
                self.add_dtype(*x)

    def create_learner(self):
        '''
        The "learner" uses knowledge of the task goals to determine the "intended"
        action of the BMI subject and pairs this intention estimation with actual observations.
        '''
        from . import clda
        self.learn_flag = False
        self.learner = clda.DumbLearner()


    def create_updater(self):
        '''
        The "updater" uses the output batches of data from the learner and an update rule to
        alter the decoder parameters to better match the intention estimates.
        '''
        from . import clda
        self.updater = None

    def call_decoder(self, neural_obs, target_state, **kwargs):
        '''
        Run the decoder computations

        Parameters
        ----------
        neural_obs : object, typically np.array of shape (n_features, n_subbins)
            n_features is the number of neural features the decoder is expecting to decode from.
            n_subbins is the number of simultaneous observations which will be decoded (typically 1)
        target_state : np.array of shape (n_states, 1)
            The current optimal state to be in to accomplish the task. In this function call, this getsget_target_BMI_state
            used when adapting the decoder using CLDA
        kwargs : optional keyword arguments
            Optional arguments to CLDA, assist, etc.
        '''
        # Get the decoder output
        decoder_output, update_flag = self.bmi_system(neural_obs, target_state, self.state, learn_flag=self.learn_flag, **kwargs)

        self.task_data['update_bmi'] = int(update_flag)

        return decoder_output

    def get_features(self):
        '''
        Run the feature extractor to get any new features to be decoded. Called by move_plant
        '''
        start_time = self.get_time()
        return self.extractor(start_time)

    def move_plant(self, **kwargs):
        '''
        The main functions to retrieve raw observations from the neural data source and convert them to movement of the plant

        Parameters
        ----------
        **kwargs : optional keyword arguments
            optional arguments for the decoder, assist, CLDA, etc. fed to the BMISystem

        Returns
        -------
        decoder_state : np.mat
            (N, 1) vector representing the state decoded by the BMI
        '''

        # Run the feature extractor
        feature_data = self.get_features()

        # Save the "neural features" (e.g., spike counts vector) to HDF file
        for key, val in list(feature_data.items()):
            self.task_data[key] = val

        # Determine the target_state and save to file
        current_assist_level = self.get_current_assist_level()
        if np.any(current_assist_level > 0) or self.learn_flag:
            target_state = self.get_target_BMI_state(self.decoder.states)
        else:
            target_state = np.ones([self.decoder.n_states, self.decoder.n_subbins]) * np.nan


        # Determine the assistive control inputs to the Decoder
        if np.any(current_assist_level) > 0:
            current_state = self.get_current_state()

            if target_state.shape[1] > 1:
                assist_kwargs = self.assister(current_state, target_state[:,0].reshape(-1,1), current_assist_level, mode=self.state)
            else:
                assist_kwargs = self.assister(current_state, target_state, current_assist_level, mode=self.state)

            kwargs.update(assist_kwargs)

        # Run the decoder
        if self.state not in self.static_states:
            neural_features = feature_data[self.extractor.feature_type]

            tmp = self.call_decoder(neural_features, target_state, **kwargs)
            self.task_data['internal_decoder_state'] = tmp

        # Drive the plant to the decoded state, if permitted by the constraints of the plant
        # If not possible, plant.drive should also take care of setting the decoder's
        # state as close as possible to physical reality
        self.plant.drive(self.decoder)
        try:
            self.dec_cnt += 1
        except:
            self.dec_cnt = 0

        self.task_data['decoder_state'] = decoder_state = self.decoder.get_state(shape=(-1,1))
        return decoder_state

    def get_current_assist_level(self):
        return self.current_assist_level

    def get_current_state(self):
        '''
        In most cases, the current state of the plant needed for calculating assistive control inputs will be stored in the decoder
        '''
        return self.decoder.filt.state.mean

    def get_target_BMI_state(self, *args):
        '''
        Run the goal calculator to determine what the target state of the task is.
        Since this is not a real task, this function must be
        overridden in child classes if any of the assist/CLDA functionality is to be used.
        '''
        raise NotImplementedError

    def _cycle(self):
        self.move_plant()
        # print(self.decoder.filt.C)
        # save loop time to HDF file
        self.task_data['loop_time'] = self.iter_time()
        super(BMILoop, self)._cycle()

    def enable_clda(self):
        print("CLDA enabled")
        self.learn_flag = True

    def disable_clda(self):
        print(("CLDA disabled after %d successful trials" % self.calc_state_occurrences('reward')))
        self.learn_flag = False

    def cleanup_hdf(self):
        '''
        Re-open the HDF file and save any extra task data kept in RAM
        '''
        super(BMILoop, self).cleanup_hdf()
        log_file = open(os.path.join(os.getenv("HOME"), 'code/bmi3d/log/clda_log'), 'w')
        log_file.write(str(self.state) + '\n')
        try:
            from . import clda
            if len(self.bmi_system.param_hist) > 0 and not self.updater is None:
                log_file.write('n_updates: %g\n' % len(self.bmi_system.param_hist))
                ignore_none = self.learner.batch_size > 1
                log_file.write('Ignoring "None" values: %s\n' % str(ignore_none))

                self.write_clda_data_to_hdf_table(
                    self.h5file.name, self.bmi_system.param_hist,
                    ignore_none=ignore_none)
        except:
            import traceback
            traceback.print_exc(file=log_file)
        log_file.close()

    @staticmethod
    def write_clda_data_to_hdf_table(hdf_fname, data, ignore_none=False):
        '''
        Save CLDA data generated during the experiment to the specified HDF file

        Parameters
        ----------
        hdf_fname : string
            filename of HDF file
        data : list
            list of dictionaries with the same keys and same dtypes for values

        Returns
        -------
        None
        '''
        log_file = open(os.path.expandvars('$HOME/code/bmi3d/log/clda_hdf_log'), 'w')

        compfilt = tables.Filters(complevel=5, complib="zlib", shuffle=True)
        if len(data) > 0:
            # Find the first parameter update dictionary
            k = 0
            first_update = data[k]
            while first_update is None:
                k += 1
                first_update = data[k]

            table_col_names = list(first_update.keys())
            print(table_col_names)
            dtype = []
            shapes = []
            for col_name in table_col_names:
                if isinstance(first_update[col_name], float):
                    shape = (1,)
                else:
                    shape = first_update[col_name].shape
                dtype.append((col_name.replace('.', '_'), 'f8', shape))
                shapes.append(shape)

            log_file.write(str(dtype))
            # Create the HDF table with the datatype above
            dtype = np.dtype(dtype)

            h5file = tables.open_file(hdf_fname, mode='a')
            arr = h5file.create_table("/", 'clda', dtype, filters=compfilt)

            null_update = np.zeros((1,), dtype=dtype)
            for col_name in table_col_names:
                null_update[col_name.replace('.', '_')] *= np.nan

            for k, param_update in enumerate(data):
                log_file.write('%d, %s\n' % (k, str(ignore_none)))
                if param_update == None:
                    if ignore_none:
                        continue
                    else:
                        data_row = null_update
                else:
                    data_row = np.zeros((1,), dtype=dtype)
                    for col_name in table_col_names:
                        data_row[col_name.replace('.', '_')] = np.asarray(param_update[col_name])

                arr.append(data_row)
            h5file.close()

    def cleanup(self, database, saveid, **kwargs):
        super(BMILoop, self).cleanup(database, saveid, **kwargs)

        # Resave decoder with drift-parameter saved as prev_task_drift_corr:
        if hasattr(self.decoder.filt, 'drift_corr'):
            print(('saving decoder: ', self.decoder.filt.drift_corr, self.decoder.filt.prev_drift_corr))
            decoder_name = self.decoder.name + '_d'+str(saveid)
            decoder_tempfilename = self.decoder.save()

            # Link the pickled decoder file to the associated task entry in the database
            dbname = kwargs['dbname'] if 'dbname' in kwargs else 'default'
            if dbname == 'default':
                database.save_bmi(decoder_name, saveid, decoder_tempfilename)
            else:
                database.save_bmi(decoder_name, saveid, decoder_tempfilename, dbname=dbname)

        # Open a log file in case of error b/c errors not visible to console
        # at this point
        log_path = os.path.join(os.path.dirname(__file__), '../../log')
        f = open(os.path.join(log_path, 'clda_cleanup_log'), 'w')
        f.write('Opening log file\n')

        f.write('# of paramter updates: %d\n' % len(self.bmi_system.param_hist))

        # save out the parameter history and new decoder unless task was stopped
        # before 1st update
        try:
            if len(self.bmi_system.param_hist) > 0  and not self.updater is None:
                # create name for new decoder
                now = datetime.datetime.now()
                decoder_name = self.decoder_sequence + now.strftime('%m%d%H%M')

                # Pickle the decoder
                # filename = self.decoder.save()

                # from db.tracker import dbq
                # suffix = f"saved_from_{self.saveid}"
                # dbq.save_bmi(suffix, self.saveid, filename)
                decoder_tempfilename = self.decoder.save()

                # Link the pickled decoder file to the associated task entry in the database
                dbname = kwargs['dbname'] if 'dbname' in kwargs else 'default'
                if dbname == 'default':
                    database.save_bmi(decoder_name, saveid, decoder_tempfilename)
                else:
                    database.save_bmi(decoder_name, saveid, decoder_tempfilename, dbname=dbname)
        except:
            traceback.print_exc(file=f)
        f.close()


class BMI(object):
    '''
    Legacy class, used only for unpickling super old Decoder objects. Ignore completely.
    '''
    pass




'''
Django's standard place to look for URL pattern matching. See
https://docs.djangoproject.com/en/dev/topics/http/urls/
for more complete documentation
'''
from django.urls import path

# Uncomment the next two lines to enable the admin:
from django.contrib import admin
admin.autodiscover()

from .tracker import ajax, views, dbq

urlpatterns = [
    path('', views.main),

    path('setup', views.setup),
    path('setup/subjects', views.setup_subjects),
    path('setup/tasks', views.setup_tasks),
    path('setup/features', views.setup_features),
    path('setup/parameters', views.setup_parameters),
    path('setup/add/subject', ajax.add_new_subject),
    path('setup/add/experimenter', ajax.add_new_experimenter),
    path('setup/add/task', ajax.add_new_task),
    path('setup/add/feature', ajax.add_new_feature),
    path('setup/add/system', ajax.add_new_system),
    path('setup/populate_models', ajax.populate_models),
    path('setup/run_upkeep', ajax.setup_run_upkeep),
    path('setup/update/generic', ajax.setup_handler),
    path('setup/update/toggle_features', ajax.toggle_features),
    path('setup/remove/task', ajax.remove_task),
    path('setup/remove/system', ajax.remove_system),
    path('setup/remove/task', ajax.remove_task),
    path('setup/remove/subject', ajax.remove_subject),
    path('setup/remove/experimenter', ajax.remove_experimenter),
    
    path(r'exp_log/', views.list_exp_history, dict(max_entries=400)),
    path(r'exp_log/all/', views.list_exp_history),
    path("exp_log/link_data_files/<int:task_entry_id>", views.link_data_files_view_generator),
    path("exp_log/link_data_files/<int:task_entry_id>/submit", views.link_data_files_response_handler),
    path("exp_log/trigger_control", ajax.trigger_control),
    # path(r'listdb/(?P<dbname>.+?)/.*?/ajax/exp_info/(?P<idx>\d+)/', ajax.exp_info),
    # path(r'listdb/(?P<dbname>.+?)/.*?/ajax/task_info/(?P<idx>\d+)/', ajax.task_info),

    # path(r'listdb/(?P<dbname>.+?)/(?P<subject>.+?)/(?P<task>.+?)', views.listdb),
    # path(r'listdb/(?P<dbname>.+?)/(?P<subject>.+?)', views.listdb),
    # path(r'listdb/(?P<dbname>.+?)/', views.listdb),
    path(r'exp_log/get_status', ajax.get_status),
    path(r'exp_log/report', ajax.get_report),
    path(r'exp_log/ajax/task_info/<int:idx>/', ajax.task_info),
    path(r'exp_log/ajax/exp_info/<int:idx>/', ajax.exp_info),
    path(r'exp_log/ajax/add_sequence', ajax.add_sequence),
    path(r'exp_log/all/ajax/exp_info/<int:idx>/', ajax.exp_info),
    path(r'exp_log/ajax/hide_entry/<int:idx>', ajax.hide_entry),
    path(r'exp_log/ajax/show_entry/<int:idx>', ajax.show_entry),
    path(r'exp_log/ajax/remove_entry/<int:idx>', ajax.remove_entry),
    path(r'exp_log/ajax/backup_entry/<int:idx>', ajax.backup_entry),
    path(r'exp_log/ajax/unbackup_entry/<int:idx>', ajax.unbackup_entry),
    path(r'exp_log/ajax/template_entry/<int:idx>', ajax.template_entry),
    path(r'exp_log/ajax/untemplate_entry/<int:idx>', ajax.untemplate_entry),
    path(r'ajax/gen_info/<int:idx>/', ajax.gen_info),
    path(r'exp_log/ajax/save_notes/<int:idx>/', ajax.save_notes),
    path(r'exp_log/all/ajax/save_notes/<int:idx>/', ajax.save_notes),
    path(r'make_bmi/<int:idx>', ajax.train_decoder_ajax_handler),
    path(r'ajax/setattr/<str:attr>/<str:value>', ajax.set_task_attr),

    path(r'start', ajax.start_experiment),
    path(r'test', ajax.start_experiment, dict(save=False)),
    path(r'saverec', ajax.start_experiment, dict(save=True, execute=False)),
    path(r'exp_log/save_entry_name', ajax.save_entry_name),

    path(r'exp_log/stop/', ajax.stop_experiment),
    path(r'enable_clda/', ajax.enable_clda),
    path(r'rewarddrain/<str:onoff>/', ajax.reward_drain),
    path(r'disable_clda/', ajax.disable_clda),
    path(r'sequence_for/<int:idx>/', views.get_sequence),
    path(r'RPC2/', dbq.rpc_handler),
    # Uncomment the next line to enable the admin:
    path(r'admin/', admin.site.urls),
]

